from . import exceptions


class Filter:

    def __init__(self):
        self.payload = None


class Timescale(Filter):
    """Filter which changes the speed and pitch of a track. Do be warned that this filter is bugged as of the lastest
        Lavalink dev version due to the filter patch not corresponding with the track time. In short this means that your
        track will either end prematurely or end later due to this. This is not the library's fault.
    """

    def __init__(self, *, speed: float = 1.0, pitch: float = 1.0, rate: float = 1.0):
        super().__init__()

        if speed < 0:
            raise exceptions.FilterInvalidArgument("Timescale speed must be more than 0.")
        if pitch < 0:
            raise exceptions.FilterInvalidArgument("Timescale pitch must be more than 0.")
        if rate < 0:
            raise exceptions.FilterInvalidArgument("Timescale rate must be more than 0.")

        self.speed = speed
        self.pitch = pitch
        self.rate = rate

        self.payload = {"timescale": {"speed": self.speed,
                                      "pitch": self.pitch,
                                      "rate": self.rate}}

    def __repr__(self):
        return f"<Pomice.TimescaleFilter speed={self.speed} pitch={self.pitch} rate={self.rate}>"


class Karaoke(Filter):
    """
    Filter which filters the vocal track from any song and leaves the instrumental. Best for karaoke as the filter implies.
    """

    def __init__(self, *, level: float, mono_level: float, filter_band: float, filter_width: float):
        super().__init__()

        self.level = level
        self.mono_level = mono_level
        self.filter_band = filter_band
        self.filter_width = filter_width

        self.payload = {"karaoke": {"level": self.level,
                                    "monoLevel": self.mono_level,
                                    "filterBand": self.filter_band,
                                    "filterWidth": self.filter_width}}

    def __repr__(self):
        return f"<Pomice.KaraokeFilter level={self.level} mono_level={self.mono_level} filter_band={self.filter_band} filter_width={self.filter_width}>"


class Tremolo(Filter):
    """Filter which produces a wavering tone in the music, causing it to sound like the music is changing in volume rapidly."""

    def __init__(self, *, frequency: float, depth: float):
        super().__init__()

        if frequency < 0:
            raise exceptions.FilterInvalidArgument("Tremolo frequency must be more than 0.")
        if depth < 0 or depth > 1:
            raise exceptions.FilterInvalidArgument("Tremolo depth must be between 0 and 1.")

        self.frequency = frequency
        self.depth = depth

        self.payload = {"tremolo": {"frequency": self.frequency,
                                    "depth": self.depth}}

    def __repr__(self):
        return f"<Pomice.TremoloFilter frequency={self.frequency} depth={self.depth}>"


class Vibrato(Filter):
    """Filter which produces a wavering tone in the music, similar to the Tremolo filter, but changes in pitch rather than volume."""

    def __init__(self, *, frequency: float, depth: float):

        super().__init__()
        if frequency < 0 or frequency > 14:
            raise exceptions.FilterInvalidArgument("Vibrato frequency must be between 0 and 14.")
        if depth < 0 or depth > 1:
            raise exceptions.FilterInvalidArgument("Vibrato depth must be between 0 and 1.")

        self.frequency = frequency
        self.depth = depth

        self.payload = {"vibrato": {"frequency": self.frequency,
                                    "depth": self.depth}}

    def __repr__(self):
        return f"<Pomice.VibratoFilter frequency={self.frequency} depth={self.depth}>"


from typing import Sequence, Union, List

from ..oauth import set_required_scopes
from . import SpotifyBase
from .track import Track
from .album import Album


class Library(SpotifyBase):
    """A Spotify Users Library.

    Attributes
    ----------
    user : :class:`Spotify.User`
        The user which this library object belongs to.
    """

    def __init__(self, client, user):
        self.user = user
        self.__client = client

    def __repr__(self):
        return f"<spotify.Library: {self.user!r}>"

    def __eq__(self, other):
        return type(self) is type(other) and self.user == other.user

    def __ne__(self, other):
        return not self.__eq__(other)

    @set_required_scopes("user-library-read")
    async def contains_albums(self, *albums: Sequence[Union[str, Album]]) -> List[bool]:
        """Check if one or more albums is already saved in the current Spotify user’s ‘Your Music’ library.

        Parameters
        ----------
        albums : Union[Album, str]
            A sequence of artist objects or spotify IDs
        """
        _albums = [str(obj) for obj in albums]
        return await self.user.http.is_saved_album(_albums)

    @set_required_scopes("user-library-read")
    async def contains_tracks(self, *tracks: Sequence[Union[str, Track]]) -> List[bool]:
        """Check if one or more tracks is already saved in the current Spotify user’s ‘Your Music’ library.

        Parameters
        ----------
        tracks : Union[Track, str]
            A sequence of track objects or spotify IDs
        """
        _tracks = [str(obj) for obj in tracks]
        return await self.user.http.is_saved_track(_tracks)

    @set_required_scopes("user-library-read")
    async def get_tracks(self, *, limit=20, offset=0) -> List[Track]:
        """Get a list of the songs saved in the current Spotify user’s ‘Your Music’ library.

        Parameters
        ----------
        limit : Optional[int]
            The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optional[int]
            The index of the first item to return. Default: 0
        """
        data = await self.user.http.saved_tracks(limit=limit, offset=offset)

        return [Track(self.__client, item["track"]) for item in data["items"]]

    @set_required_scopes("user-library-read")
    async def get_all_tracks(self) -> List[Track]:
        """Get a list of all the songs saved in the current Spotify user’s ‘Your Music’ library.

        Returns
        -------
        tracks : List[:class:`Track`]
            The tracks of the artist.
        """
        tracks: List[Track] = []
        total = None
        offset = 0

        while True:
            data = await self.user.http.saved_tracks(limit=50, offset=offset)

            if total is None:
                total = data["total"]

            offset += 50
            tracks += list(
                Track(self.__client, item["track"]) for item in data["items"]
            )

            if len(tracks) >= total:
                break

        return tracks

    @set_required_scopes("user-library-read")
    async def get_albums(self, *, limit=20, offset=0) -> List[Album]:
        """Get a list of the albums saved in the current Spotify user’s ‘Your Music’ library.

        Parameters
        ----------
        limit : Optional[int]
            The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optional[int]
            The index of the first item to return. Default: 0
        """
        data = await self.user.http.saved_albums(limit=limit, offset=offset)

        return [Album(self.__client, item["album"]) for item in data["items"]]

    @set_required_scopes("user-library-read")
    async def get_all_albums(self) -> List[Album]:
        """Get a list of the albums saved in the current Spotify user’s ‘Your Music’ library.

        Returns
        -------
        albums : List[:class:`Album`]
            The albums.
        """
        albums: List[Album] = []
        total = None
        offset = 0

        while True:
            data = await self.user.http.saved_albums(limit=50, offset=offset)

            if total is None:
                total = data["total"]

            offset += 50
            albums += list(
                Album(self.__client, item["album"]) for item in data["items"]
            )

            if len(albums) >= total:
                break

        return albums

    @set_required_scopes("user-library-modify")
    async def remove_albums(self, *albums):
        """Remove one or more albums from the current user’s ‘Your Music’ library.

        Parameters
        ----------
        albums : Sequence[Union[Album, str]]
            A sequence of artist objects or spotify IDs
        """
        _albums = [(obj if isinstance(obj, str) else obj.id) for obj in albums]
        await self.user.http.delete_saved_albums(",".join(_albums))

    @set_required_scopes("user-library-modify")
    async def remove_tracks(self, *tracks):
        """Remove one or more tracks from the current user’s ‘Your Music’ library.

        Parameters
        ----------
        tracks : Sequence[Union[Track, str]]
            A sequence of track objects or spotify IDs
        """
        _tracks = [(obj if isinstance(obj, str) else obj.id) for obj in tracks]
        await self.user.http.delete_saved_tracks(",".join(_tracks))

    @set_required_scopes("user-library-modify")
    async def save_albums(self, *albums):
        """Save one or more albums to the current user’s ‘Your Music’ library.

        Parameters
        ----------
        albums : Sequence[Union[Album, str]]
            A sequence of artist objects or spotify IDs
        """
        _albums = [(obj if isinstance(obj, str) else obj.id) for obj in albums]
        await self.user.http.save_albums(",".join(_albums))

    @set_required_scopes("user-library-modify")
    async def save_tracks(self, *tracks):
        """Save one or more tracks to the current user’s ‘Your Music’ library.

        Parameters
        ----------
        tracks : Sequence[Union[Track, str]]
            A sequence of track objects or spotify IDs
        """
        _tracks = [(obj if isinstance(obj, str) else obj.id) for obj in tracks]
        await self.user.http.save_tracks(_tracks)



from .pool import NodePool


class PomiceEvent:
    """The base class for all events dispatched by a node. 
        Every event must be formatted within your bots code as a listener.
        i.e: If you want to listen for when a track starts, the event would be:
        ```py
        @bot.listen
        async def on_pomice_track_start(self, event):
        ```
    """
    def __init__(self):
        pass 

    name = 'event'

class TrackStartEvent(PomiceEvent):
    """Fired when a track has successfully started. Returns the player associated with said track and the track ID"""
    def __init__(self, data):
        super().__init__()

        self.name = "track_start"
        self.player = NodePool.get_node().get_player(int(data["guildId"]))
        self.track_id = data['track']

    def __repr__(self) -> str:
        return f"<Pomice.TrackStartEvent track_id={self.track_id}>"

        

class TrackEndEvent(PomiceEvent):
    """Fired when a track has successfully ended. Returns the player associated with the track along with the track ID and reason."""
    def __init__(self, data):
        super().__init__()

        self.name = "track_end"
        self.player = NodePool.get_node().get_player(int(data["guildId"]))
        self.track_id = data['track']
        self.reason = data['reason']

    def __repr__(self) -> str:
        return f"<Pomice.TrackEndEvent track_id={self.track_id} reason={self.reason}>"

class TrackStuckEvent(PomiceEvent):
    """Fired when a track is stuck and cannot be played. Returns the player associated with the track along with a track ID to be further parsed by the end user."""
    def __init__(self, data):
        super().__init__()

        self.name = "track_stuck"
        self.player = NodePool.get_node().get_player(int(data["guildId"]))

        self.track_id = data["track"]
        self.threshold = data["thresholdMs"]

    def __repr__(self) -> str:
        return f"<Pomice.TrackStuckEvent track_id={self.track_id} threshold={self.threshold}>"

class TrackExceptionEvent(PomiceEvent):
    """Fired when a track error has occured. Returns the player associated with the track along with the error code and execption"""
    def __init__(self, data):
        super().__init__()

        self.name = "track_exception"
        self.player = NodePool.get_node().get_player(int(data["guildId"]))

        self.error = data["error"]
        self.exception = data["exception"]

    def __repr__(self) -> str:
        return f"<Pomice.TrackExceptionEvent> error={self.error} exeception={self.exception}"

class WebsocketClosedEvent(PomiceEvent):
    """Fired when a websocket connection to a node has been closed. Returns the reason and the error code."""
    def __init__(self, data):
        super().__init__()

        self.name = "websocket_closed"

        self.reason = data["reason"]
        self.code = data["code"]

    def __repr__(self) -> str:
        return f"<Pomice.WebsocketClosedEvent reason={self.reason} code={self.code}>"

class WebsocketOpenEvent(PomiceEvent):
    """Fired when a websocket connection to a node has been initiated. Returns the target and the session SSRC."""
    def __init__(self, data):
        super().__init__()

        self.name = "websocket_open"
        
        self.target: str = data['target']
        self.ssrc: int = data['ssrc']

    def __repr__(self) -> str:
        return f"<Pomice.WebsocketOpenEvent target={self.target} ssrc={self.ssrc}>"



from discord.ext import commands

class Track:
    """
    The base track object. Returns critical track information needed to be parsed by Lavalink.
    You can also pass in commands.Context to get a discord.py Context object by passing in a valid Context object when you search for a track.
    """

    def __init__(self, track_id: str, info: dict, ctx: commands.Context = None):

        self.track_id = track_id
        self.info = info

        self.title = info.get("title")
        self.author = info.get("author")
        self.length = info.get("length")
        if ctx:
            self.ctx: commands.Context = ctx
            self.requester = self.ctx.author
        self.identifier = info.get("identifier")
        self.uri = info.get("uri")
        self.is_stream = info.get("isStream")
        self.is_seekable = info.get("isSeekable")
        self.position = info.get("position")

    def __str__(self):
        return self.title

    def __repr__(self):
        return f"<Pomice.track title={self.title!r} uri=<{self.uri!r}> length={self.length}>"


class Playlist:
    """
    The base playlist object. Returns critcal playlist information like the name of the playlist and what tracks are included to be parsed by Lavalink.
    You can also pass in commands.Context to get a discord.py Context object by passing in a valid Context object when you search for a track.
    """

    def __init__(self, playlist_info: dict, tracks: list, ctx: commands.Context = None):

        self.playlist_info = playlist_info
        self.tracks_raw = tracks

        self.name = playlist_info.get("name")
        self.selected_track = playlist_info.get("selectedTrack")

        self.tracks = [Track(track_id=track["track"], info=track["info"], ctx=ctx) for track in self.tracks_raw]
        

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<Pomice.playlist name={self.name!r} track_count={len(self.tracks)}>"


from functools import partial
from itertools import islice
from typing import List, Optional, Union, Callable, Tuple, Iterable, TYPE_CHECKING, Any, Dict, Set

from ..oauth import set_required_scopes
from ..http import HTTPUserClient, HTTPClient
from . import AsyncIterable, URIBase, Track, PlaylistTrack, Image

if TYPE_CHECKING:
    import spotify


class MutableTracks:
    __slots__ = (
        "playlist",
        "tracks",
        "was_empty",
        "is_empty",
        "replace_tracks",
        "get_all_tracks",
    )

    def __init__(self, playlist: "Playlist") -> None:
        self.playlist = playlist
        self.tracks = tracks = getattr(playlist, "_Playlist__tracks")

        if tracks is not None:
            self.was_empty = self.is_empty = not tracks

        self.replace_tracks = playlist.replace_tracks
        self.get_all_tracks = playlist.get_all_tracks

    async def __aenter__(self):
        if self.tracks is None:
            self.tracks = tracks = list(await self.get_all_tracks())
            self.was_empty = self.is_empty = not tracks
        else:
            tracks = list(self.tracks)

        return tracks

    async def __aexit__(self, typ, value, traceback):
        if self.was_empty and self.is_empty:
            # the tracks were empty and is still empty.
            # skip the api call.
            return

        tracks = self.tracks

        await self.replace_tracks(*tracks)
        setattr(self.playlist, "_Playlist__tracks", tuple(self.tracks))


class Playlist(URIBase, AsyncIterable):  # pylint: disable=too-many-instance-attributes
    """A Spotify Playlist.

    Attributes
    ----------
    collaborative : :class:`bool`
        Returns true if context is not search and the owner allows other users to modify the playlist. Otherwise returns false.
    description : :class:`str`
        The playlist description. Only returned for modified, verified playlists, otherwise null.
    url : :class:`str`
        The open.spotify URL.
    followers : :class:`int`
        The total amount of followers
    href : :class:`str`
        A link to the Web API endpoint providing full details of the playlist.
    id : :class:`str`
        The Spotify ID for the playlist.
    images : List[:class:`spotify.Image`]
        Images for the playlist.
        The array may be empty or contain up to three images.
        The images are returned by size in descending order.
        If returned, the source URL for the image ( url ) is temporary and will expire in less than a day.
    name : :class:`str`
        The name of the playlist.
    owner : :class:`spotify.User`
        The user who owns the playlist
    public : :class`bool`
        The playlist’s public/private status:
            true the playlist is public,
            false the playlist is private,
            null the playlist status is not relevant.
    snapshot_id : :class:`str`
        The version identifier for the current playlist.
    tracks : Optional[Tuple[:class:`PlaylistTrack`]]
        A tuple of :class:`PlaylistTrack` objects or `None`.
    """

    __slots__ = (
        "collaborative",
        "description",
        "url",
        "followers",
        "href",
        "id",
        "images",
        "name",
        "owner",
        "public",
        "uri",
        "total_tracks",
        "__client",
        "__http",
        "__tracks",
    )

    __tracks: Optional[Tuple[PlaylistTrack, ...]]
    __http: Union[HTTPUserClient, HTTPClient]
    total_tracks: Optional[int]

    def __init__(
        self,
        client: "spotify.Client",
        data: Union[dict, "Playlist"],
        *,
        http: Optional[HTTPClient] = None,
    ):
        self.__client = client
        self.__http = http or client.http

        assert self.__http is not None

        self.__tracks = None
        self.total_tracks = None

        if not isinstance(data, (Playlist, dict)):
            raise TypeError("data must be a Playlist instance or a dict.")

        if isinstance(data, dict):
            self.__from_raw(data)
        else:
            for name in filter((lambda name: name[0] != "_"), Playlist.__slots__):
                setattr(self, name, getattr(data, name))

        # AsyncIterable attrs
        self.__aiter_klass__ = PlaylistTrack
        self.__aiter_fetch__ = partial(
            client.http.get_playlist_tracks, self.id, limit=50
        )

    def __repr__(self):
        return f'<spotify.Playlist: {getattr(self, "name", None) or self.id}>'

    def __len__(self):
        return self.total_tracks

    # Internals

    def __from_raw(self, data: dict) -> None:
        from .user import User

        client = self.__client

        self.id = data.pop("id")  # pylint: disable=invalid-name

        self.images = tuple(Image(**image) for image in data.pop("images", []))
        self.owner = User(client, data=data.pop("owner"))

        self.public = data.pop("public")
        self.collaborative = data.pop("collaborative")
        self.description = data.pop("description", None)
        self.followers = data.pop("followers", {}).get("total", None)
        self.href = data.pop("href")
        self.name = data.pop("name")
        self.url = data.pop("external_urls").get("spotify", None)
        self.uri = data.pop("uri")

        tracks: Optional[Tuple[PlaylistTrack, ...]] = (
            tuple(PlaylistTrack(client, item) for item in data["tracks"]["items"])
            if "items" in data["tracks"]
            else None
        )

        self.__tracks = tracks

        self.total_tracks = (
            data["tracks"]["total"]
        )

    # Track retrieval

    @set_required_scopes(None)
    async def get_tracks(
        self, *, limit: Optional[int] = 20, offset: Optional[int] = 0
    ) -> Tuple[PlaylistTrack, ...]:
        """Get a fraction of a playlists tracks.

        Parameters
        ----------
        limit : Optional[int]
            The limit on how many tracks to retrieve for this playlist (default is 20).
        offset : Optional[int]
            The offset from where the api should start from in the tracks.

        Returns
        -------
        tracks : Tuple[PlaylistTrack]
            The tracks of the playlist.
        """
        data = await self.__http.get_playlist_tracks(
            self.id, limit=limit, offset=offset
        )
        return tuple(PlaylistTrack(self.__client, item) for item in data["items"])

    @set_required_scopes(None)
    async def get_all_tracks(self) -> Tuple[PlaylistTrack, ...]:
        """Get all playlist tracks from the playlist.

        Returns
        -------
        tracks : Tuple[:class:`PlaylistTrack`]
            The playlists tracks.
        """
        tracks: List[PlaylistTrack] = []
        offset = 0

        if self.total_tracks is None:
            self.total_tracks = (
                await self.__http.get_playlist_tracks(self.id, limit=1, offset=0)
            )["total"]

        while len(tracks) < self.total_tracks:
            data = await self.__http.get_playlist_tracks(
                self.id, limit=50, offset=offset
            )

            tracks += [PlaylistTrack(self.__client, item) for item in data["items"]]
            offset += 50

        self.total_tracks = len(tracks)
        return tuple(tracks)

    # Playlist structure modification

    # Basic api wrapping

    @set_required_scopes("playlist-modify-public", "playlist-modify-private")
    async def add_tracks(self, *tracks) -> str:
        """Add one or more tracks to a user’s playlist.

        Parameters
        ----------
        tracks : Iterable[Union[:class:`str`, :class:`Track`]]
            Tracks to add to the playlist

        Returns
        -------
        snapshot_id : :class:`str`
            The snapshot id of the playlist.
        """
        data = await self.__http.add_playlist_tracks(
            self.id, tracks=[str(track) for track in tracks]
        )
        return data["snapshot_id"]

    @set_required_scopes("playlist-modify-public", "playlist-modify-private")
    async def remove_tracks(
        self, *tracks: Union[str, Track, Tuple[Union[str, Track], List[int]]]
    ):
        """Remove one or more tracks from a user’s playlist.

        Parameters
        ----------
        tracks : Iterable[Union[:class:`str`, :class:`Track`]]
            Tracks to remove from the playlist

        Returns
        -------
        snapshot_id : :class:`str`
            The snapshot id of the playlist.
        """
        tracks_: List[Union[str, Dict[str, Union[str, Set[int]]]]] = []

        for part in tracks:
            if not isinstance(part, (Track, str, tuple)):
                raise TypeError(
                    "Track argument of tracks parameter must be a Track instance, string or a tuple of those and an iterator of positive integers."
                )

            if isinstance(part, (Track, str)):
                tracks_.append(str(part))
                continue

            track, positions, = part

            if not isinstance(track, (Track, str)):
                raise TypeError(
                    "Track argument of tuple track parameter must be a Track instance or a string."
                )

            if not hasattr(positions, "__iter__"):
                raise TypeError("Positions element of track tuple must be a iterator.")

            if not all(isinstance(index, int) for index in positions):
                raise TypeError("Members of the positions iterator must be integers.")

            elem: Dict[str, Union[str, Set[int]]] = {"uri": str(track), "positions": set(positions)}
            tracks_.append(elem)

        data = await self.__http.remove_playlist_tracks(self.id, tracks=tracks_)
        return data["snapshot_id"]

    @set_required_scopes("playlist-modify-public", "playlist-modify-private")
    async def replace_tracks(self, *tracks: Union[Track, PlaylistTrack, str]) -> None:
        """Replace all the tracks in a playlist, overwriting its existing tracks.

        This powerful request can be useful for replacing tracks, re-ordering existing tracks, or clearing the playlist.

        Parameters
        ----------
        tracks : Iterable[Union[:class:`str`, :class:`Track`]]
            Tracks to place in the playlist
        """
        bucket: List[str] = []
        for track in tracks:
            if not isinstance(track, (str, Track)):
                raise TypeError(
                    f"tracks must be a iterable of strings or Track instances. Got {type(track)!r}"
                )

            bucket.append(str(track))

        body: Tuple[str, ...] = tuple(bucket)

        head: Tuple[str, ...]
        tail: Tuple[str, ...]
        head, tail = body[:100], body[100:]

        if head:
            await self.__http.replace_playlist_tracks(self.id, tracks=head)

        while tail:
            head, tail = tail[:100], tail[100:]
            await self.extend(head)

    @set_required_scopes("playlist-modify-public", "playlist-modify-private")
    async def reorder_tracks(
        self,
        start: int,
        insert_before: int,
        length: int = 1,
        *,
        snapshot_id: Optional[str] = None,
    ) -> str:
        """Reorder a track or a group of tracks in a playlist.

        Parameters
        ----------
        start : int
            The position of the first track to be reordered.
        insert_before : int
            The position where the tracks should be inserted.
        length : Optional[int]
            The amount of tracks to be reordered. Defaults to 1 if not set.
        snapshot_id : str
            The playlist’s snapshot ID against which you want to make the changes.

        Returns
        -------
        snapshot_id : str
            The snapshot id of the playlist.
        """
        data = await self.__http.reorder_playlists_tracks(
            self.id, start, length, insert_before, snapshot_id=snapshot_id
        )
        return data["snapshot_id"]

    # Library functionality.

    @set_required_scopes("playlist-modify-public", "playlist-modify-private")
    async def clear(self):
        """Clear the playlists tracks.

        .. note::

            This method will mutate the current
            playlist object, and the spotify Playlist.

        .. warning::

            This is a desctructive operation and can not be reversed!
        """
        await self.__http.replace_playlist_tracks(self.id, tracks=[])

    @set_required_scopes("playlist-modify-public", "playlist-modify-private")
    async def extend(self, tracks: Union["Playlist", Iterable[Union[Track, str]]]):
        """Extend a playlists tracks with that of another playlist or a list of Track/Track URIs.

        .. note::

            This method will mutate the current
            playlist object, and the spotify Playlist.

        Parameters
        ----------
        tracks : Union["Playlist", List[Union[Track, str]]]
            Tracks to add to the playlist, acceptable values are:
             - A :class:`spotify.Playlist` object
             - A :class:`list` of :class:`spotify.Track` objects or Track URIs

        Returns
        -------
        snapshot_id : str
            The snapshot id of the playlist.
        """
        bucket: Iterable[Union[Track, str]]

        if isinstance(tracks, Playlist):
            bucket = await tracks.get_all_tracks()

        elif not hasattr(tracks, "__iter__"):
            raise TypeError(
                f"`tracks` was an invalid type, expected any of: Playlist, Iterable[Union[Track, str]], instead got {type(tracks)}"
            )

        else:
            bucket = list(tracks)

        gen: Iterable[str] = (str(track) for track in bucket)

        while True:
            head: List[str] = list(islice(gen, 0, 100))

            if not head:
                break

            await self.__http.add_playlist_tracks(self.id, tracks=head)

    @set_required_scopes("playlist-modify-public", "playlist-modify-private")
    async def insert(self, index, obj: Union[PlaylistTrack, Track]) -> None:
        """Insert an object before the index.

        .. note::

            This method will mutate the current
            playlist object, and the spotify Playlist.
        """
        if not isinstance(obj, (PlaylistTrack, Track)):
            raise TypeError(
                f"Expected a PlaylistTrack or Track object instead got {obj!r}"
            )

        async with MutableTracks(self) as tracks:
            tracks.insert(index, obj)

    @set_required_scopes("playlist-modify-public", "playlist-modify-private")
    async def pop(self, index: int = -1) -> PlaylistTrack:
        """Remove and return the track at the specified index.

        .. note::

            This method will mutate the current
            playlist object, and the spotify Playlist.

        Returns
        -------
        playlist_track : :class:`PlaylistTrack`
            The track that was removed.

        Raises
        ------
        IndexError
            If there are no tracks or the index is out of range.
        """
        async with MutableTracks(self) as tracks:
            return tracks.pop(index)

    @set_required_scopes("playlist-modify-public", "playlist-modify-private")
    async def sort(
        self,
        *,
        key: Optional[Callable[[PlaylistTrack], bool]] = None,
        reverse: Optional[bool] = False,
    ) -> None:
        """Stable sort the playlist in place.

        .. note::

            This method will mutate the current
            playlist object, and the spotify Playlist.
        """
        async with MutableTracks(self) as tracks:
            tracks.sort(key=key, reverse=reverse)

    @set_required_scopes("playlist-modify-public", "playlist-modify-private")
    async def remove(self, value: Union[PlaylistTrack, Track]) -> None:
        """Remove the first occurence of the value.

        .. note::

            This method will mutate the current
            playlist object, and the spotify Playlist.

        Raises
        -------
        ValueError
            If the value is not present.
        """
        async with MutableTracks(self) as tracks:
            tracks.remove(value)

    @set_required_scopes("playlist-modify-public", "playlist-modify-private")
    async def copy(self) -> "Playlist":
        """Return a shallow copy of the playlist object.

        Returns
        -------
        playlist : :class:`Playlist`
            The playlist object copy.
        """
        return Playlist(client=self.__client, data=self, http=self.__http)

    @set_required_scopes("playlist-modify-public", "playlist-modify-private")
    async def reverse(self) -> None:
        """Reverse the playlist in place.

        .. note::

            This method will mutate the current
            playlist object, and the spotify Playlist.
        """
        async with MutableTracks(self) as tracks:
            tracks.reverse()



class PomiceException(Exception):
    """Base of all Pomice exceptions."""


class NodeException(Exception):
    """Base exception for nodes."""


class NodeCreationError(NodeException):
    """There was a problem while creating the node."""


class NodeConnectionFailure(NodeException):
    """There was a problem while connecting to the node."""


class NodeConnectionClosed(NodeException):
    """The nodes connection is closed."""
    pass


class NodeNotAvailable(PomiceException):
    """The node is not currently available."""
    pass


class NoNodesAvailable(PomiceException):
    """There are no nodes currently available."""
    pass


class TrackInvalidPosition(PomiceException):
    """An invalid position was chosen for a track."""
    pass


class TrackLoadError(PomiceException):
    """There was an error while loading a track."""
    pass


class FilterInvalidArgument(PomiceException):
    """An invalid argument was passed to a filter."""
    pass

class SpotifyAlbumLoadFailed(PomiceException):
    """The pomice Spotify client was unable to load an album"""
    pass

class SpotifyTrackLoadFailed(PomiceException):
    """The pomice Spotify client was unable to load a track"""
    pass

class SpotifyPlaylistLoadFailed(PomiceException):
    """The pomice Spotify client was unable to load a playlist"""
    pass

class InvalidSpotifyClientAuthorization(PomiceException):
    """No Spotify client authorization was provided in order to use the Spotify track search feature"""
    pass


from typing import Union, Optional, List

from ..oauth import set_required_scopes
from . import SpotifyBase, Device, Track
from .typing import SomeURIs, SomeURI

Offset = Union[int, str, Track]
SomeDevice = Union[Device, str]


class Player(SpotifyBase):  # pylint: disable=too-many-instance-attributes
    """A Spotify Users current playback.

    Attributes
    ----------
    device : :class:`spotify.Device`
        The device that is currently active.
    repeat_state : :class:`str`
        "off", "track", "context"
    shuffle_state : :class:`bool`
        If shuffle is on or off.
    is_playing : :class:`bool`
        If something is currently playing.
    """

    def __init__(self, client, user, data):
        self.__client = client
        self.__user = user

        self.repeat_state = data.get("repeat_state", None)
        self.shuffle_state = data.pop("shuffle_state", None)
        self.is_playing = data.pop("is_playing", None)
        self.device = Device(data=data.pop("device", None))

    def __repr__(self):
        return f"<spotify.Player: {self.user!r}>"

    # Properties

    @property
    def user(self):
        return self.__user

    # Public methods

    @set_required_scopes("user-modify-playback-state")
    async def pause(self, *, device: Optional[SomeDevice] = None):
        """Pause playback on the user’s account.

        Parameters
        ----------
        device : Optional[:obj:`SomeDevice`]
            The Device object or id of the device this command is targeting.
            If not supplied, the user’s currently active device is the target.
        """
        device_id: Optional[str] = str(device) if device is not None else None
        await self.user.http.pause_playback(device_id=device_id)

    @set_required_scopes("user-modify-playback-state")
    async def resume(self, *, device: Optional[SomeDevice] = None):
        """Resume playback on the user's account.

        Parameters
        ----------
        device : Optional[:obj:`SomeDevice`]
            The Device object or id of the device this command is targeting.
            If not supplied, the user’s currently active device is the target.
        """
        device_id: Optional[str] = str(device) if device is not None else None
        await self.user.http.play_playback(None, device_id=device_id)

    @set_required_scopes("user-modify-playback-state")
    async def seek(self, pos, *, device: Optional[SomeDevice] = None):
        """Seeks to the given position in the user’s currently playing track.

        Parameters
        ----------
        pos : int
            The position in milliseconds to seek to.
            Must be a positive number.
            Passing in a position that is greater than the length of the track will cause the player to start playing the next song.
        device : Optional[:obj:`SomeDevice`]
            The Device object or id of the device this command is targeting.
            If not supplied, the user’s currently active device is the target.
        """
        device_id: Optional[str] = str(device) if device is not None else None
        await self.user.http.seek_playback(pos, device_id=device_id)

    @set_required_scopes("user-modify-playback-state")
    async def set_repeat(self, state, *, device: Optional[SomeDevice] = None):
        """Set the repeat mode for the user’s playback.

        Parameters
        ----------
        state : str
            Options are repeat-track, repeat-context, and off
        device : Optional[:obj:`SomeDevice`]
            The Device object or id of the device this command is targeting.
            If not supplied, the user’s currently active device is the target.
        """
        device_id: Optional[str] = str(device) if device is not None else None
        await self.user.http.repeat_playback(state, device_id=device_id)

    @set_required_scopes("user-modify-playback-state")
    async def set_volume(self, volume: int, *, device: Optional[SomeDevice] = None):
        """Set the volume for the user’s current playback device.

        Parameters
        ----------
        volume : int
            The volume to set. Must be a value from 0 to 100 inclusive.
        device : Optional[:obj:`SomeDevice`]
            The Device object or id of the device this command is targeting.
            If not supplied, the user’s currently active device is the target.
        """
        device_id: Optional[str] = str(device) if device is not None else None
        await self.user.http.set_playback_volume(volume, device_id=device_id)

    @set_required_scopes("user-modify-playback-state")
    async def next(self, *, device: Optional[SomeDevice] = None):
        """Skips to next track in the user’s queue.

        Parameters
        ----------
        device : Optional[:obj:`SomeDevice`]
            The Device object or id of the device this command is targeting.
            If not supplied, the user’s currently active device is the target.
        """
        device_id: Optional[str] = str(device) if device is not None else None
        await self.user.http.skip_next(device_id=device_id)

    @set_required_scopes("user-modify-playback-state")
    async def previous(self, *, device: Optional[SomeDevice] = None):
        """Skips to previous track in the user’s queue.

        Note that this will ALWAYS skip to the previous track, regardless of the current track’s progress.
        Returning to the start of the current track should be performed using :meth:`seek`

        Parameters
        ----------
        device : Optional[:obj:`SomeDevice`]
            The Device object or id of the device this command is targeting.
            If not supplied, the user’s currently active device is the target.
        """
        device_id: Optional[str] = str(device) if device is not None else None
        return await self.user.http.skip_previous(device_id=device_id)

    @set_required_scopes("user-modify-playback-state")
    async def enqueue(self, uri: SomeURI, device: Optional[SomeDevice] = None):
        """Add an item to the end of the user’s current playback queue.

        Parameters
        ----------
        uri : Union[:class:`spotify.URIBase`, :class:`str`]
            The uri of the item to add to the queue. Must be a track or an
            episode uri.
        device_id : Optional[Union[Device, :class:`str`]]
            The id of the device this command is targeting. If not supplied,
            the user’s currently active device is the target.
        """
        device_id: Optional[str]
        if device is not None:
            if not isinstance(device, (Device, str)):
                raise TypeError(
                    f"Expected `device` to either be a spotify.Device or a string. got {type(device)!r}"
                )

            device_id = str(device)
        else:
            device_id = None

        await self.user.http.playback_queue(uri=str(uri), device_id=device_id)

    @set_required_scopes("user-modify-playback-state")
    async def play(
        self,
        *uris: SomeURIs,
        offset: Optional[Offset] = 0,
        device: Optional[SomeDevice] = None,
    ):
        """Start a new context or resume current playback on the user’s active device.

        The method treats a single argument as a Spotify context, such as a Artist, Album and playlist objects/URI.
        When called with multiple positional arguments they are interpreted as a array of Spotify Track objects/URIs.

        Parameters
        ----------
        *uris : SomeURI
            When a single argument is passed in that argument is treated
            as a context (except if it is a track or track uri).
            Valid contexts are: albums, artists, playlists.
            Album, Artist and Playlist objects are accepted too.
            Otherwise when multiple arguments are passed in they,
            A sequence of Spotify Tracks or Track URIs to play.
        offset : Optional[:obj:`Offset`]
            Indicates from where in the context playback should start.
            Only available when `context` corresponds to an album or playlist object,
            or when the `uris` parameter is used. when an integer offset is zero based and can’t be negative.
        device : Optional[:obj:`SomeDevice`]
            The Device object or id of the device this command is targeting.
            If not supplied, the user’s currently active device is the target.
        """
        context_uri: Union[List[str], str]

        if (
            len(uris) > 1
            or isinstance(uris[0], Track)
            or (isinstance(uris[0], str) and "track" in uris[0])
        ):
            # Regular uris paramter
            context_uri = [str(uri) for uri in uris]
        else:
            # Treat it as a context URI
            context_uri = str(uris[0])

        device_id: Optional[str]
        if device is not None:
            if not isinstance(device, (Device, str)):
                raise TypeError(
                    f"Expected `device` to either be a spotify.Device or a string. got {type(device)!r}"
                )

            device_id = str(device)
        else:
            device_id = None

        await self.user.http.play_playback(
            context_uri, offset=offset, device_id=device_id
        )

    @set_required_scopes("user-modify-playback-state")
    async def shuffle(
        self, state: Optional[bool] = None, *, device: Optional[SomeDevice] = None
    ):
        """shuffle on or off for user’s playback.

        Parameters
        ----------
        state : Optional[bool]
            if `True` then Shuffle user’s playback.
            else if `False` do not shuffle user’s playback.
        device : Optional[:obj:`SomeDevice`]
            The Device object or id of the device this command is targeting.
            If not supplied, the user’s currently active device is the target.
        """
        device_id: Optional[str] = str(device) if device is not None else None
        await self.user.http.shuffle_playback(state, device_id=device_id)

    @set_required_scopes("user-modify-playback-state")
    async def transfer(self, device: SomeDevice, ensure_playback: bool = False):
        """Transfer playback to a new device and determine if it should start playing.

        Parameters
        ----------
        device : :obj:`SomeDevice`
            The device on which playback should be started/transferred.
        ensure_playback : bool
            if `True` ensure playback happens on new device.
            else keep the current playback state.
        """
        device_id: Optional[str] = str(device) if device is not None else None
        await self.user.http.transfer_player(device_id=device_id, play=ensure_playback)



"""Pomice wrapper for Lavalink, made possible by cloudwithax 2021"""

__version__ = "1.0.3"
__title__ = "pomice"
__author__ = "cloudwithax"


from .exceptions import *
from .events import *
from .filters import *
from .objects import *
from .pool import NodePool
from .node import Node
from .player import Player


import time

import discord

from . import exceptions
from . import filters
from . import objects
from .node import Node
from .pool import NodePool
from . import events

import discord
from discord import VoiceProtocol, VoiceChannel
from discord.ext import commands
from typing import Dict, Optional, Any, Union



class Player(VoiceProtocol):
    """The base player class for Pomice. In order to initiate a player, you must pass it in as a cls when you connect to a channel.
    i.e: ```py
            await ctx.author.voice.channel.connect(cls=pomice.Player)```
    """

    def __init__(self, client: Union[commands.Bot, discord.Client, commands.AutoShardedBot, discord.AutoShardedClient], channel: VoiceChannel):
        super().__init__(client=client, channel=channel)

        self.client: Union[commands.Bot, discord.Client, commands.AutoShardedBot, discord.AutoShardedClient] = client
        self._bot: Union[commands.Bot, discord.Client, commands.AutoShardedBot, discord.AutoShardedClient] = client
        self.channel: VoiceChannel = channel
        self._guild: discord.Guild = channel.guild
        self._dj: discord.Member = None

        self._node: Node = NodePool.get_node()
        self._current: objects.Track = None
        self._filter: filters.Filter = None
        self._volume: int = 100
        self._paused: bool = False
        self._is_connected: bool = False

        self._position: int = 0
        self._last_update: int = 0
        self._current_track_id = None


        self._session_id: Optional[str] = None
        self._voice_server_update_data: Optional[dict] = None


    def __repr__(self):
        return f"<Pomice.player bot={self._bot} guildId={self._guild.id} is_connected={self.is_connected} is_playing={self.is_playing}>"

    @property
    def position(self):
        """Property which returns the player's position in a track in milliseconds"""

        if not self.is_playing or not self._current:
            return 0

        if self.is_paused:
            return min(self._last_position, self._current.length)

        difference = (time.time() * 1000) - self._last_update
        position = self._last_position + difference

        if position > self.current.length:
            return 0

        return min(position, self.current.length)

    @property
    def is_connected(self):
        """Property which returns whether or not the player is connected to a node."""
        return self._is_connected

    @property
    def is_playing(self):
        """Property which returns whether or not the player is actively playing a track."""
        return self._is_connected and self._current is not None

    @property
    def is_paused(self):
        """Property which returns whether or not the player has a track which is paused or not."""
        return self._is_connected and self._paused is True

    @property
    def node(self):
        """Property which returns what node is associated with this player."""
        return self._node

    @property
    def current(self):
        """Property which returns the current track as a Pomice Track object"""
        return self._current

    @property
    def volume(self):
        """Property which returns the players current volume as an integer"""
        return self._volume


    async def _update_state(self, data: dict):

        state: dict = data.get('state')
        self._last_update = time.time() * 1000
        self._is_connected = state.get('connected')
        self._last_position = state.get('position')

    async def _dispatch_voice_update(self) -> None:

        if not self._session_id or not self._voice_server_update_data:
            return

        await self._node.send(op='voiceUpdate', sessionId=self._session_id, guildId=str(self._guild.id), event={**self._voice_server_update_data})

    async def _voice_server_update(self, data: dict):

        self._voice_server_update_data = data
        await self._dispatch_voice_update()

        
    async def _voice_state_update(self, data: dict):

        if not (channel_id := data.get('channel_id')):
            self.channel, self._session_id, self._voice_server_update_data = None
            return

        self.channel = self._guild.get_channel(int(channel_id))
        self._session_id = data.get('session_id')
        await self._dispatch_voice_update()

    async def _dispatch_event(self, data: dict):
        event_type = data.get('type')
        event = getattr(events, event_type, None)
        event = event(data)
        self._bot.dispatch(f"pomice_{event.name}", event)

    async def get_tracks(self, query: str, ctx: commands.Context = None):
        """Fetches tracks from the node's REST api to parse into Lavalink.
        If you passed in Spotify API credentials, you can also pass in a Spotify URL of a playlist, album or track
        and it will be parsed accordingly.
        You can also pass in a discord.py Context object to get a Context object on any track you search.
        """
        return await self._node.get_tracks(query, ctx)

    async def connect(self, *, timeout: float, reconnect: bool):
        await self._guild.change_voice_state(channel=self.channel)
        self._node._players[self._guild.id] = self
        self._is_connected = True
    
    async def stop(self):
        """Stops a currently playing track."""
        self._current = None
        await self._node.send(op='stop', guildId=str(self._guild.id))

    async def disconnect(self, *, force: bool = False):
        await self.stop()
        await self._guild.change_voice_state(channel=None)
        self.cleanup()
        self.channel = None
        self._is_connected = False
        del self._node._players[self._guild.id]

    async def destroy(self):
        """Disconnects a player and destroys the player instance."""
        await self.disconnect() 
        await self._node.send(op='destroy', guildId=str(self._guild.id))

    async def play(self, track: objects.Track, start_position: int = 0):
        """Plays a track. If a Spotify track is passed in, it will be handled accordingly."""
        if track.track_id == "spotify":
            track: objects.Track = await self._node.get_tracks(f"{track.title} {track.author}")
        await self._node.send(op='play', guildId=str(self._guild.id), track=track.track_id, startTime=start_position, endTime=track.length, noReplace=False)
        self._current = track
        return self._current

    async def seek(self, position: float):
        """Seeks to a position in the currently playing track milliseconds"""

        if position < 0 or position > self.current.length:
            raise exceptions.TrackInvalidPosition(f"Seek position must be between 0 and the track length")

        await self._node.send(op='seek', guildId=str(self._guild.id), position=position)
        return self.position

    async def set_pause(self, pause: bool):
        """Sets the pause state of the currently playing track."""
        await self._node.send(op='pause', guildId=str(self._guild.id), pause=pause)
        self._paused = pause
        return self._paused

    async def set_volume(self, volume: int):
        """Sets the volume of the player as an integer. Lavalink accepts an amount from 0 to 500.""" 
        await self._node.send(op='volume', guildId=str(self._guild.id), volume=volume)
        self._volume = volume
        return self._volume

    async def set_filter(self, filter: filters.Filter):
        """Sets a filter of the player. Takes a pomice.Filter object. This will only work if you are using the development version of Lavalink."""
        await self._node.send(op='filters', guildId=str(self._guild.id), **filter.payload)
        await self.seek(self.position)
        self._filter = filter
        return filter

  


from functools import partial
from typing import Optional, List, TYPE_CHECKING

from ..oauth import set_required_scopes
from . import AsyncIterable, URIBase, Image

if TYPE_CHECKING:
    import spotify


class Artist(URIBase, AsyncIterable):  # pylint: disable=too-many-instance-attributes
    """A Spotify Artist.

    Attributes
    ----------
    id : str
        The Spotify ID of the artist.
    uri : str
        The URI of the artist.
    url : str
        The open.spotify URL.
    href : str
        A link to the Web API endpoint providing full details of the artist.
    name : str
        The name of the artist.
    genres : List[str]
        A list of the genres the artist is associated with.
        For example: "Prog Rock" , "Post-Grunge". (If not yet classified, the array is empty.)
    followers : Optional[int]
        The total number of followers.
    popularity : int
        The popularity of the artist.
        The value will be between 0 and 100, with 100 being the most popular.
        The artist’s popularity is calculated from the popularity of all the artist’s tracks.
    images : List[Image]
        Images of the artist in various sizes, widest first.
    """

    def __init__(self, client, data):
        self.__client = client

        # Simplified object attributes
        self.id = data.pop("id")  # pylint: disable=invalid-name
        self.uri = data.pop("uri")
        self.url = data.pop("external_urls").get("spotify", None)
        self.href = data.pop("href")
        self.name = data.pop("name")

        # Full object attributes
        self.genres = data.pop("genres", None)
        self.followers = data.pop("followers", {}).get("total", None)
        self.popularity = data.pop("popularity", None)
        self.images = list(Image(**image) for image in data.pop("images", []))

        # AsyncIterable attrs
        from .album import Album

        self.__aiter_klass__ = Album
        self.__aiter_fetch__ = partial(
            self.__client.http.artist_albums, self.id, limit=50
        )

    def __repr__(self):
        return f"<spotify.Artist: {self.name!r}>"

    # Public

    @set_required_scopes(None)
    async def get_albums(
        self,
        *,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0,
        include_groups=None,
        market: Optional[str] = None,
    ) -> List["spotify.Album"]:
        """Get the albums of a Spotify artist.

        Parameters
        ----------
        limit : Optional[int]
            The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optiona[int]
            The offset of which Spotify should start yielding from.
        include_groups : INCLUDE_GROUPS_TP
            INCLUDE_GROUPS
        market : Optional[str]
            An ISO 3166-1 alpha-2 country code.

        Returns
        -------
        albums : List[Album]
            The albums of the artist.
        """
        from .album import Album

        data = await self.__client.http.artist_albums(
            self.id,
            limit=limit,
            offset=offset,
            include_groups=include_groups,
            market=market,
        )
        return list(Album(self.__client, item) for item in data["items"])

    @set_required_scopes(None)
    async def get_all_albums(self, *, market="US") -> List["spotify.Album"]:
        """loads all of the artists albums, depending on how many the artist has this may be a long operation.

        Parameters
        ----------
        market : Optional[str]
            An ISO 3166-1 alpha-2 country code.

        Returns
        -------
        albums : List[Album]
            The albums of the artist.
        """
        from .album import Album

        albums: List[Album] = []
        offset = 0
        total = await self.total_albums(market=market)

        while len(albums) < total:
            data = await self.__client.http.artist_albums(
                self.id, limit=50, offset=offset, market=market
            )

            offset += 50
            albums += list(Album(self.__client, item) for item in data["items"])

        return albums

    @set_required_scopes(None)
    async def total_albums(self, *, market: str = None) -> int:
        """get the total amout of tracks in the album.

        Parameters
        ----------
        market : Optional[str]
            An ISO 3166-1 alpha-2 country code.

        Returns
        -------
        total : int
            The total amount of albums.
        """
        data = await self.__client.http.artist_albums(
            self.id, limit=1, offset=0, market=market
        )
        return data["total"]

    @set_required_scopes(None)
    async def top_tracks(self, country: str = "US") -> List["spotify.Track"]:
        """Get Spotify catalog information about an artist’s top tracks by country.

        Parameters
        ----------
        country : str
            The country to search for, it defaults to 'US'.

        Returns
        -------
        tracks : List[Track]
            The artists top tracks.
        """
        from .track import Track

        top = await self.__client.http.artist_top_tracks(self.id, country=country)
        return list(Track(self.__client, item) for item in top["tracks"])

    @set_required_scopes(None)
    async def related_artists(self) -> List["Artist"]:
        """Get Spotify catalog information about artists similar to a given artist.

        Similarity is based on analysis of the Spotify community’s listening history.

        Returns
        -------
        artists : List[Artist]
            The artists deemed similar.
        """
        related = await self.__client.http.artist_related_artists(self.id)
        return list(Artist(self.__client, item) for item in related["artists"])



# pylint: skip-file

import asyncio
import sys
import json
import time
from typing import (
    Optional,
    List,
    Sequence,
    Union,
    Dict,
    Awaitable,
    BinaryIO,
    Tuple,
    Any,
)
from base64 import b64encode
from urllib.parse import quote

import aiohttp
import backoff  # type: ignore

from . import __version__
from .utils import filter_items
from .errors import (
    HTTPException,
    Forbidden,
    NotFound,
    SpotifyException,
    BearerTokenError,
    RateLimitedException,
)

__all__ = ("HTTPClient", "HTTPUserClient")

_GET_BEARER_ARG_ERR = "{name} was `None` when getting a bearer token."
_PYTHON_VERSION = ".".join(str(_) for _ in sys.version_info[:3])
_AIOHTTP_VERSION = aiohttp.__version__


class HTTPClient:
    """A class responsible for handling all HTTP logic.

    This class combines a small amount of stateful logic control
    with the :meth:`request` method and a very thin wrapper over
    the raw HTTP API.

    All endpoint methods mirror the default arguments the API
    uses and is best described as a series of "good defaults"
    for the routes.

    Parameters
    ----------
    client_id : str
        The client id provided by spotify for the app.
    client_secret : str
        The client secret for the app.
    loop : Optional[event loop]
        The event loop the client should run on, if no loop is specified `asyncio.get_event_loop()` is called and used instead.


    Attributes
    ----------
    loop : AbstractEventLoop
        The loop the client is running with.
    client_id : str
        The client id of the app.
    client_secret : str
        The client secret.
    """

    RETRY_AMOUNT = 10
    DEFAULT_USER_AGENT = (
        user_agent
    ) = f"Application (https://github.com/mental32/spotify.py {__version__}) Python/{_PYTHON_VERSION} aiohttp/{_AIOHTTP_VERSION}"

    def __init__(self, client_id: str, client_secret: str, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self._session = aiohttp.ClientSession(loop=self.loop)

        self.client_id = client_id
        self.client_secret = client_secret

        self.bearer_info: Optional[Dict[str, str]] = None

        self.__request_barrier_lock = asyncio.Lock()
        self.__request_barrier = asyncio.Event()
        self.__request_barrier.set()

    @staticmethod
    def route(
        method: str, path: str, *, base: str = "https://api.spotify.com/v1", **kwargs
    ) -> Tuple[str, str]:
        """Used for constructing URLs for API endpoints.

        Parameters
        ----------
        method : str
            The HTTP/REST method used.
        path : str
            A path to be formatted.
        kwargs : Any
            The arguments used to format the path.

        Returns
        -------
        route : Tuple[str, str]
            A tuple of the method and formatted url path to use.
        """
        url = base + path

        if kwargs:
            url = url.format(
                **{
                    key: (quote(value) if isinstance(value, str) else value)
                    for key, value in kwargs.items()
                }
            )

        return (method, url)

    async def get_bearer_info(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        """Get the application bearer token from client_id and client_secret.

        Raises
        ------
        SpotifyException
            This will be raised when either `client_id` or
            `client_secret` is `None`
        """
        client_id = client_id or self.client_id
        client_secret = client_secret or self.client_secret

        if client_id is None:
            raise SpotifyException(_GET_BEARER_ARG_ERR.format(name="client_id"))

        if client_secret is None:
            raise SpotifyException(_GET_BEARER_ARG_ERR.format(name="client_secret"))

        token = b64encode(":".join((client_id, client_secret)).encode())

        data = {"grant_type": "client_credentials"}
        headers = {"Authorization": f"Basic {token.decode()}"}

        session = session or self._session

        async with session.post(
            "https://accounts.spotify.com/api/token", data=data, headers=headers
        ) as response:
            bearer_info = json.loads(await response.text(encoding="utf-8"))

            if "error" in bearer_info.keys():
                raise BearerTokenError(response=response, message=bearer_info)

        return bearer_info

    @backoff.on_exception(backoff.expo, RateLimitedException)
    async def request(self, route, **kwargs):
        r"""Make a request to the spotify API with the current bearer credentials.

        Parameters
        ----------
        route : Tuple[str, str]
            A tuple of the method and url gained from :meth:`route`.
        \*\*kwargs : Any
            keyword arguments to pass into :class:`aiohttp.ClientSession.request`
        """
        assert isinstance(route, tuple), "route parameter was not a tuple!"
        assert len(route) == 2, "route parameter must have exactly two items"

        method, url, = route

        headers = kwargs.pop("headers", {})
        if "Authorization" not in headers:
            if self.bearer_info is None:
                self.bearer_info = bearer_info = await self.get_bearer_info()
                access_token = bearer_info["access_token"]
            else:
                access_token = self.bearer_info["access_token"]

            headers["Authorization"] = "Bearer " + access_token

        headers = {
            "Content-Type": kwargs.pop("content_type", "application/json"),
            "User-Agent": self.user_agent,
            **headers,
        }

        if "json" in kwargs:
            headers["Content-Type"] = "application/json"
            kwargs["data"] = json.dumps(
                kwargs.pop("json"), separators=(",", ":"), ensure_ascii=True
            )

        for current_retry in range(self.RETRY_AMOUNT):
            await self.__request_barrier.wait()

            response = await self._session.request(
                method, url, headers=headers, **kwargs
            )

            try:
                status = response.status

                try:
                    data = json.loads(await response.text(encoding="utf-8"))
                except json.decoder.JSONDecodeError:
                    data = {}

                if 300 > status >= 200:
                    return data

                if status == 401:
                    self.bearer_info = bearer_info = await self.get_bearer_info()
                    headers["Authorization"] = "Bearer " + bearer_info["access_token"]
                    continue

                if status == 429:
                    # we're being rate limited.

                    self.__request_barrier.clear()
                    amount = int(response.headers.get("Retry-After"))
                    checkpoint = int(time.time())

                    async with self.__request_barrier_lock:
                        if (int(time.time()) - checkpoint) < amount:
                            self.__request_barrier.clear()
                            await asyncio.sleep(int(amount), loop=self.loop)
                            self.__request_barrier.set()

                    continue

                if status in (502, 503):
                    # unconditional retry
                    continue

                if status == 403:
                    raise Forbidden(response, data)

                if status == 404:
                    raise NotFound(response, data)
            finally:
                await response.release()

        if response.status == 429:
            raise RateLimitedException((amount, _max_retries - current_retry))

        raise HTTPException(response, data)

    async def close(self):
        """Close the underlying HTTP session."""
        await self._session.close()

    # Methods are defined in the order that they are listed in
    # the api docs (https://developer.spotify.com/documentation/web-api/reference/)

    # Album related endpoints

    def album(self, spotify_id: str, market: Optional[str] = "US") -> Awaitable:
        """Get Spotify catalog information for a single album.

        Parameters
        ----------
        spotify_id : str
            The spotify_id to search by.
        market : Optional[str]
            An ISO 3166-1 alpha-2 country code.
        """
        route = self.route("GET", "/albums/{spotify_id}", spotify_id=spotify_id)
        payload: Dict[str, Any] = {}

        if market:
            payload["market"] = market

        return self.request(route, params=payload)

    def album_tracks(
        self,
        spotify_id: str,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0,
        market="US",
    ) -> Awaitable:
        """Get Spotify catalog information about an album’s tracks.

        Parameters
        ----------
        spotify_id : str
            The spotify_id to search by.
        limit : Optional[int]
            The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optiona[int]
            The offset of which Spotify should start yielding from.
        market : Optional[str]
            An ISO 3166-1 alpha-2 country code.
        """
        route = self.route("GET", "/albums/{spotify_id}/tracks", spotify_id=spotify_id)
        payload: Dict[str, Any] = {"limit": limit, "offset": offset}

        if market:
            payload["market"] = market

        return self.request(route, params=payload)

    def albums(self, spotify_ids, market="US") -> Awaitable:
        """Get Spotify catalog information for multiple albums identified by their Spotify IDs.

        Parameters
        ----------
        spotify_ids : List[str]
            The spotify_ids to search by.
        market : Optional[str]
            An ISO 3166-1 alpha-2 country code.
        """
        route = self.route("GET", "/albums/")
        payload: Dict[str, Any] = {"ids": spotify_ids}

        if market:
            payload["market"] = market

        return self.request(route, params=payload)

    # Artist related endpoints.

    def artist(self, spotify_id) -> Awaitable:
        """Get Spotify catalog information for a single artist identified by their unique Spotify ID.

        Parameters
        ----------
        spotify_id : str
            The spotify_id to search by.
        """
        route = self.route("GET", "/artists/{spotify_id}", spotify_id=spotify_id)
        return self.request(route)

    def artist_albums(
        self,
        spotify_id,
        include_groups=None,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0,
        market="US",
    ):
        """Get Spotify catalog information about an artist’s albums.

        Parameters
        ----------
        spotify_id : str
            The spotify_id to search by.
        include_groups : INCLUDE_GROUPS_TP
            INCLUDE_GROUPS
        limit : Optional[int]
            The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optiona[int]
            The offset of which Spotify should start yielding from.
        market : Optional[str]
            An ISO 3166-1 alpha-2 country code.
        """
        route = self.route("GET", "/artists/{spotify_id}/albums", spotify_id=spotify_id)
        payload: Dict[str, Any] = {"limit": limit, "offset": offset}

        if include_groups:
            payload["include_groups"] = include_groups

        if market:
            payload["market"] = market

        return self.request(route, params=payload)

    def artist_top_tracks(self, spotify_id, country) -> Awaitable:
        """Get Spotify catalog information about an artist’s top tracks by country.

        Parameters
        ----------
        spotify_id : str
            The spotify_id to search by.
        country : COUNTRY_TP
            COUNTRY
        """
        route = self.route(
            "GET", "/artists/{spotify_id}/top-tracks", spotify_id=spotify_id
        )
        payload: Dict[str, Any] = {"country": country}
        return self.request(route, params=payload)

    def artist_related_artists(self, spotify_id) -> Awaitable:
        """Get Spotify catalog information about artists similar to a given artist.

        Similarity is based on analysis of the Spotify community’s listening history.

        Parameters
        ----------
        spotify_id : str
            The spotify_id to search by.
        """
        route = self.route(
            "GET", "/artists/{spotify_id}/related-artists", spotify_id=spotify_id
        )
        return self.request(route)

    def artists(self, spotify_ids) -> Awaitable:
        """Get Spotify catalog information for several artists based on their Spotify IDs.

        Parameters
        ----------
        spotify_id : List[str]
            The spotify_ids to search with.
        """
        route = self.route("GET", "/artists")
        payload: Dict[str, Any] = {"ids": spotify_ids}
        return self.request(route, params=payload)

    # Browse endpoints.

    def category(self, category_id, country=None, locale=None) -> Awaitable:
        """Get a single category used to tag items in Spotify (on, for example, the Spotify player’s “Browse” tab).

        Parameters
        ----------
        category_id : str
            The Spotify category ID for the category.
        country : COUNTRY_TP
            COUNTRY
        locale : LOCALE_TP
            LOCALE
        """
        route = self.route(
            "GET", "/browse/categories/{category_id}", category_id=category_id
        )
        payload: Dict[str, Any] = {}

        if country:
            payload["country"] = country

        if locale:
            payload["locale"] = locale

        return self.request(route, params=payload)

    def category_playlists(
        self,
        category_id,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0,
        country=None,
    ) -> Awaitable:
        """Get a list of Spotify playlists tagged with a particular category.

        Parameters
        ----------
        category_id : str
            The Spotify category ID for the category.
        limit : Optional[int]
            The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optional[int]
            The index of the first item to return. Default: 0
        country : COUNTRY_TP
            COUNTRY
        """
        route = self.route(
            "GET", "/browse/categories/{category_id}/playlists", category_id=category_id
        )
        payload: Dict[str, Any] = {"limit": limit, "offset": offset}

        if country:
            payload["country"] = country

        return self.request(route, params=payload)

    def categories(
        self,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0,
        country=None,
        locale=None,
    ) -> Awaitable:
        """Get a list of categories used to tag items in Spotify (on, for example, the Spotify player’s “Browse” tab).

        Parameters
        ----------
        limit : Optional[int]
            The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optional[int]
            The index of the first item to return. Default: 0
        country : COUNTRY_TP
            COUNTRY
        locale : LOCALE_TP
            LOCALE
        """
        route = self.route("GET", "/browse/categories")
        payload: Dict[str, Any] = {"limit": limit, "offset": offset}

        if country:
            payload["country"] = country

        if locale:
            payload["locale"] = locale

        return self.request(route, params=payload)

    def featured_playlists(
        self,
        locale=None,
        country=None,
        timestamp=None,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0,
    ):
        """Get a list of Spotify featured playlists (shown, for example, on a Spotify player’s ‘Browse’ tab).

        Parameters
        ----------
        locale : LOCALE_TP
            LOCALE
        country : COUNTRY_TP
            COUNTRY
        timestamp : TIMESTAMP_TP
            TIMESTAMP
        limit : Optional[int]
            The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optional[int]
            The index of the first item to return. Default: 0
        """
        route = self.route("GET", "/browse/featured-playlists")
        payload: Dict[str, Any] = {"limit": limit, "offset": offset}

        if country:
            payload["country"] = country

        if locale:
            payload["locale"] = locale

        if timestamp:
            payload["timestamp"] = timestamp

        return self.request(route, params=payload)

    def new_releases(
        self, *, country=None, limit: Optional[int] = 20, offset: Optional[int] = 0
    ) -> Awaitable:
        """Get a list of new album releases featured in Spotify (shown, for example, on a Spotify player’s “Browse” tab).

        Parameters
        ----------
        limit : Optional[int]
            The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optional[int]
            The index of the first item to return. Default: 0
        country : COUNTRY_TP
            COUNTRY
        """
        route = self.route("GET", "/browse/new-releases")
        payload: Dict[str, Any] = {"limit": limit, "offset": offset}

        if country:
            payload["country"] = country

        return self.request(route, params=payload)

    def recommendations(
        self,
        seed_artists,
        seed_genres,
        seed_tracks,
        *,
        limit: Optional[int] = 20,
        market=None,
        **filters,
    ):
        """Get Recommendations Based on Seeds.

        Parameters
        ----------
        seed_artists : str
            A comma separated list of Spotify IDs for seed artists. Up to 5 seed values may be provided.
        seed_genres : str
            A comma separated list of any genres in the set of available genre seeds. Up to 5 seed values may be provided.
        seed_tracks : str
            A comma separated list of Spotify IDs for a seed track. Up to 5 seed values may be provided.
        limit : Optional[int]
            The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.
        market : Optional[str]
            An ISO 3166-1 alpha-2 country code.
        max_* : Optional[Keyword arguments]
            For each tunable track attribute, a hard ceiling on the selected track attribute’s value can be provided.
        min_* : Optional[Keyword arguments]
            For each tunable track attribute, a hard floor on the selected track attribute’s value can be provided.
        target_* : Optional[Keyword arguments]
            For each of the tunable track attributes (below) a target value may be provided.
        """
        route = self.route("GET", "/recommendations")
        payload: Dict[str, Any] = {
            "seed_artists": seed_artists,
            "seed_genres": seed_genres,
            "seed_tracks": seed_tracks,
            "limit": limit,
        }

        if market:
            payload["market"] = market

        if filters:
            payload.update(filters)

        return self.request(route, params=payload)

    # Follow related endpoints.

    def following_artists_or_users(self, ids, *, type_="artist") -> Awaitable:
        """Check to see if the current user is following one or more artists or other Spotify users.

        Parameters
        ----------
        ids : List[:class:`str`]
            A comma-separated list of the artist or the user Spotify IDs to check.
            A maximum of 50 IDs can be sent in one request.
        type_ : Optional[:class:`str`]
            The ID type: either "artist" or "user".
            Default: "artist"
        """
        route = self.route("GET", "/me/following/contains")
        payload: Dict[str, Any] = {"ids": ids, "type": type_}

        return self.request(route, params=payload)

    def following_playlists(self, playlist_id: str, ids: List[str]) -> Awaitable:
        """Check to see if one or more Spotify users are following a specified playlist.

        Parameters
        ----------
        playlist_id : :class:`str`
            The Spotify ID of the playlist.
        ids : List[:class:`str`]
            A list of the artist or the user Spotify IDs.
            A maximum of five IDs are allowed.
        """
        route = self.route(
            "GET",
            "/playlists/{playlist_id}/followers/contains",
            playlist_id=playlist_id,
        )
        payload: Dict[str, Any] = {"ids": ids}

        return self.request(route, params=payload)

    def follow_artist_or_user(self, type_: str, ids: List[str]) -> Awaitable:
        """Add the current user as a follower of one or more artists or other Spotify users.

        Parameters
        ----------
        type_ : :class:`str`
            either artist or user.
        ids : List[:class:`str`]
            A list of the artist or the user Spotify IDs.
        """
        route = self.route("PUT", "/me/following")
        payload: Dict[str, Any] = {"ids": ids, "type": type_}

        return self.request(route, params=payload)

    def followed_artists(
        self, *, limit: Optional[int] = 20, after: Optional[str] = None
    ) -> Awaitable:
        """Get the current user’s followed artists.

        Parameters
        ----------
        limit : Optional[:class:`int`]
            The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.
        after : Optional[:class:`str`]
            The last artist ID retrieved from the previous request.
        """
        route = self.route("GET", "/me/following")
        payload: Dict[str, Any] = {"limit": limit, "type": "artist"}

        if after:
            payload["after"] = after

        return self.request(route, params=payload)

    def unfollow_artists_or_users(self, type_: str, ids: List[str]) -> Awaitable:
        """Remove the current user as a follower of one or more artists or other Spotify users.

        Parameters
        ----------
        type_ : :class:`str`
            either artist or user.
        ids : List[:class:`str`]
            A list of the artist or the user Spotify IDs.
        """
        route = self.route("DELETE", "/me/following")
        payload: Dict[str, Any] = {"ids": ids, "type": type_}

        return self.request(route, params=payload)

    def unfollow_playlist(self, playlist_id: str) -> Awaitable:
        """Remove the current user as a follower of a playlist.

        Parameters
        ----------
        playlist_id : :class:`str`
            The Spotify ID of the playlist that is to be no longer followed.
        """
        route = self.route(
            "DELETE", "/playlists/{playlist_id}/followers", playlist_id=playlist_id
        )

        return self.request(route)

    def is_saved_album(self, ids) -> Awaitable:
        """Check if one or more albums is already saved in the current Spotify user’s ‘Your Music’ library.

        Parameters
        ----------
        ids : List[:class:`str`]
            A list of the Spotify IDs.
        """
        route = self.route("GET", "/me/albums/contains")
        payload: Dict[str, Any] = {"ids": ",".join(ids)}

        return self.request(route, params=payload)

    def is_saved_track(self, ids: List[str]) -> Awaitable:
        """Check if one or more tracks is already saved in the current Spotify user’s ‘Your Music’ library.

        Parameters
        ----------
        ids : List[:class:`str`]
            A list of the Spotify IDs.
        """
        route = self.route("GET", "/me/tracks/contains")
        payload: Dict[str, Any] = {"ids": ",".join(ids)}

        return self.request(route, params=payload)

    def saved_albums(
        self,
        *,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0,
        market: Optional[str] = None,
    ) -> Awaitable:
        """Get a list of the albums saved in the current Spotify user’s ‘Your Music’ library.

        Parameters
        ----------
        limit : Optional[:class:`str`]
            The maximum number of objects to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optional[:class:`str`]
            The index of the first object to return. Default: 0 (i.e., the first object). Use with limit to get the next set of objects.
        market : Optional[:class:`str`]
            An ISO 3166-1 alpha-2 country code or the string from_token. Provide this parameter if you want to apply Track Relinking.
        """
        route = self.route("GET", "/me/albums")
        payload: Dict[str, Any] = {"limit": limit, "offset": offset}

        if market is not None:
            payload["market"] = market

        return self.request(route, params=payload)

    def saved_tracks(
        self,
        *,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0,
        market: Optional[str] = None,
    ) -> Awaitable:
        """Get a list of the songs saved in the current Spotify user’s ‘Your Music’ library.

        Parameters
        ----------
        limit : Optional[:class:`str`]
            The maximum number of objects to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optional[:class:`str`]
            The index of the first object to return. Default: 0 (i.e., the first object). Use with limit to get the next set of objects.
        market : Optional[:class:`str`]
            An ISO 3166-1 alpha-2 country code or the string from_token. Provide this parameter if you want to apply Track Relinking.
        """
        route = self.route("GET", "/me/tracks")
        payload: Dict[str, Any] = {"limit": limit, "offset": offset}

        if market:
            payload["market"] = market

        return self.request(route, params=payload)

    def delete_saved_albums(self, ids: List[str]) -> Awaitable:
        """Remove one or more albums from the current user’s ‘Your Music’ library.

        Parameters
        ----------
        ids : List[:class:`str`]
            A list of the Spotify IDs.
        """
        route = self.route("DELETE", "/me/albums")
        return self.request(route, json=ids)

    def delete_saved_tracks(self, ids: List[str]) -> Awaitable:
        """Remove one or more tracks from the current user’s ‘Your Music’ library.

        Parameters
        ----------
        ids : List[:class:`str`]
            A list of the Spotify IDs.
        """
        route = self.route("DELETE", "/me/tracks")
        return self.request(route, json=ids)

    def save_tracks(self, ids: List[str]) -> Awaitable:
        """Save one or more tracks to the current user’s ‘Your Music’ library.

        Parameters
        ----------
        ids : List[:class:`str`]
            A list of the Spotify IDs.
        """
        route = self.route("PUT", "/me/tracks")
        return self.request(route, json=ids)

    def save_albums(self, ids: List[str]) -> Awaitable:
        """Save one or more albums to the current user’s ‘Your Music’ library.

        Parameters
        ----------
        ids : List[:class:`str`]
            A list of the Spotify IDs.
        """
        route = self.route("PUT", "/me/albums")
        return self.request(route, json=ids)

    def top_artists_or_tracks(
        self,
        type_: str,
        *,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0,
        time_range: Optional[str] = None,
    ) -> Awaitable:
        """Get the current user’s top artists or tracks based on calculated affinity.

        Affinity is a measure of the expected preference a user has for a particular track or artist.
        It is based on user behavior, including play history, but does not include actions made while in incognito mode.
        Light or infrequent users of Spotify may not have sufficient play history to generate a full affinity data set.

        As a user’s behavior is likely to shift over time, this preference data is available over three time spans.

        For each time range, the top 50 tracks and artists are available for each user.
        In the future, it is likely that this restriction will be relaxed. This data is typically updated once each day for each user.

        Parameters
        ----------
        type_ : :class;`str`
            The type of entity to return. Valid values: "artists" or "tracks".
        limit : Optional[:class:`int`]
            The number of entities to return. Default: 20. Minimum: 1. Maximum: 50. For example: limit=2
        offset : Optional[:class:`int`]
            The index of the first entity to return. Default: 0 (i.e., the first track). Use with limit to get the next set of entities.
        time_range : Optional[:class:`str`]
            Over what time frame the affinities are computed.
            Valid values:
            - "long_term" (calculated from several years of data and including all new data as it becomes available)
            - "medium_term" (approximately last 6 months)
            - "short_term" (approximately last 4 weeks). Default: medium_term.
        """
        route = self.route("GET", "/me/top/{type_}", type_=type_)
        payload: Dict[str, Any] = {"limit": limit, "offset": offset}

        if time_range is not None:
            payload["time_range"] = time_range

        return self.request(route, params=payload)

    def available_devices(self) -> Awaitable:
        """Get information about a user’s available devices."""
        route = self.route("GET", "/me/player/devices")
        return self.request(route)

    def current_player(self, *, market: Optional[str] = None) -> Awaitable:
        """Get information about the user’s current playback state, including track, track progress, and active device.

        Parameters
        ----------
        market : Optional[:class:`str`]
            An ISO 3166-1 alpha-2 country code or the string from_token. Provide this parameter if you want to apply Track Relinking.
        """
        route = self.route("GET", "/me/player")
        payload: Dict[str, Any] = {}

        if market:
            payload["market"] = market

        return self.request(route, params=payload)

    def playback_queue(self, *, uri: str, device_id: Optional[str] = None) -> Awaitable:
        """Add an item to the end of the user’s current playback queue.

        Parameters
        ----------
        uri : :class:`str`
            The uri of the item to add to the queue. Must be a track or an
            episode uri.
        device_id : :class:`str`
            The id of the device this command is targeting. If not supplied,
            the user’s currently active device is the target.
        """
        route = self.route("POST", "/me/player/queue")
        params = {"uri": uri}

        if device_id is not None:
            params["device_id"] = device_id

        return self.request(route, params=params)

    def recently_played(
        self,
        *,
        limit: Optional[int] = 20,
        before: Optional[str] = None,
        after: Optional[str] = None,
    ) -> Awaitable:
        """Get tracks from the current user’s recently played tracks.

        Returns the most recent 50 tracks played by a user.
        Note that a track currently playing will not be visible in play history until it has completed.
        A track must be played for more than 30 seconds to be included in play history.

        Any tracks listened to while the user had “Private Session” enabled in their client will not be returned in the list of recently played tracks.

        The endpoint uses a bidirectional cursor for paging.
        Follow the next field with the before parameter to move back in time, or use the after parameter to move forward in time.
        If you supply no before or after parameter, the endpoint will return the most recently played songs, and the next link will page back in time.

        Parameters
        ----------
        limit : Optional[:class:`int`]
            The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.
        after : Optional[:class:`str`]
            A Unix timestamp in milliseconds. Returns all items after (but not including) this cursor position. If after is specified, before must not be specified.
        before : Optional[:class:`str`]
            A Unix timestamp in milliseconds. Returns all items before (but not including) this cursor position. If before is specified, after must not be specified.
        """
        route = self.route("GET", "/me/player/recently-played")
        payload: Dict[str, Any] = {"limit": limit}

        if before:
            payload["before"] = before
        elif after:
            payload["after"] = after

        return self.request(route, params=payload)

    def currently_playing(self, *, market: Optional[str] = None) -> Awaitable:
        """Get the object currently being played on the user’s Spotify account.

        Parameters
        ----------
        market : Optional[:class:`str`]
            An ISO 3166-1 alpha-2 country code or the string from_token. Provide this parameter if you want to apply Track Relinking.
        """
        route = self.route("GET", "/me/player/currently-playing")
        payload: Dict[str, Any] = {}

        if market:
            payload["market"] = market

        return self.request(route, params=payload)

    def pause_playback(self, *, device_id: Optional[str] = None) -> Awaitable:
        """Pause playback on the user’s account.

        Parameters
        ----------
        device_id : Optional[:class:`str`]
            The id of the device this command is targeting. If not supplied, the user’s currently active device is the target.
        """
        route = self.route("PUT", "/me/player/pause")
        payload: Dict[str, Any] = {}

        if device_id:
            payload["device_id"] = device_id

        return self.request(route, params=payload)

    def seek_playback(
        self, position_ms: int, *, device_id: Optional[str] = None
    ) -> Awaitable:
        """Seeks to the given position in the user’s currently playing track.

        Parameters
        ----------
        position_ms : :class:`int`
            The position in milliseconds to seek to. Must be a positive number. Passing in a position that is greater than the length of the track will cause the player to start playing the next song.
        device_id : Optional[:class:`str`]
            The id of the device this command is targeting. If not supplied, the user’s currently active device is the target.
        """
        route = self.route("PUT", "/me/player/seek")
        payload: Dict[str, Any] = {"position_ms": position_ms}

        if device_id:
            payload["device_id"] = device_id

        return self.request(route, params=payload)

    def repeat_playback(
        self, state: str, *, device_id: Optional[str] = None
    ) -> Awaitable:
        """Set the repeat mode for the user’s playback. Options are repeat-track, repeat-context, and off.

        Parameters
        ----------
        state : :class:`str`
            "track", "context" or "off".
             - track will repeat the current track.
             - context will repeat the current context.
             - off will turn repeat off.
        device_id : Optional[str]
            The id of the device this command is targeting. If not supplied, the user’s currently active device is the target.
        """
        route = self.route("PUT", "/me/player/repeat")
        payload: Dict[str, Any] = {"state": state}

        if device_id:
            payload["device_id"] = device_id

        return self.request(route, params=payload)

    def set_playback_volume(
        self, volume: int, *, device_id: Optional[str] = None
    ) -> Awaitable:
        """Set the volume for the user’s current playback device.

        Parameters
        ----------
        volume : :class:`int`
            The volume to set. Must be a value from 0 to 100 inclusive.
        device_id : Optional[:class:`str`]
            The id of the device this command is targeting. If not supplied, the user’s currently active device is the target.
        """
        route = self.route("PUT", "/me/player/volume")
        payload: Dict[str, Any] = {"volume_percent": volume}

        if device_id:
            payload["device_id"] = device_id

        return self.request(route, params=payload)

    def skip_next(self, *, device_id: Optional[str] = None) -> Awaitable:
        """Skips to next track in the user’s queue.

        Parameters
        ----------
        device_id : Optional[:class:`str`]
            The id of the device this command is targeting. If not supplied, the user’s currently active device is the target.
        """
        route = self.route("POST", "/me/player/next")
        payload: Dict[str, Any] = {}

        if device_id:
            payload["device_id"] = device_id

        return self.request(route, params=payload)

    def skip_previous(self, *, device_id: Optional[str] = None) -> Awaitable:
        """Skips to previous track in the user’s queue.

        Parameters
        ----------
        device_id : Optional[:class:`str`]
            The id of the device this command is targeting. If not supplied, the user’s currently active device is the target.
        """
        route = self.route("POST", "/me/player/previous")
        payload: Dict[str, Any] = {}

        if device_id:
            payload["device_id"] = device_id

        return self.request(route, params=payload)

    def play_playback(
        self,
        context_uri: Union[str, Sequence[str]],
        *,
        offset: Optional[Union[str, int]] = None,
        device_id: Optional[str] = None,
        position_ms: Optional[int] = 0,
    ) -> Awaitable:
        """Start a new context or resume current playback on the user’s active device.

        .. note::

            In order to resume playback set the context_uri to None.

        Parameters
        ----------
        context_uri : Union[str, Sequence[:class:`str`]]
            The context to play, if it is a string
            then it must be a uri of a album, artist
            or playlist.

            Otherwise a sequece of strings can be passed
            in and they must all be track URIs
        offset : Optional[Union[:class:`str`, :class:`int`]]
            The offset of which to start from,
            must either be an integer or a track URI.
        device_id : Optional[:class:`str`]
            The id of the device this command is targeting. If not supplied, the user’s currently active device is the target.
        position_ms : Optional[:class:`int`]
            indicates from what position to start playback. Must be a positive number.
            Passing in a position that is greater than the length of the track will cause the player to start playing the next song.
        """
        route = self.route("PUT", "/me/player/play")
        payload: Dict[str, Any] = {"position_ms": position_ms}
        params: Dict[str, Any] = {}
        can_set_offset: bool = False

        if isinstance(context_uri, str):
            payload["context_uri"] = context_uri
            can_set_offset = "playlist" in context_uri or "album" in context_uri

        elif hasattr(context_uri, "__iter__"):
            payload["uris"] = list(context_uri)
            can_set_offset = True

        elif context_uri is None:
            pass  # Do nothing, context_uri == None is allowed and intended for resume's

        else:
            raise TypeError(
                f"`context_uri` must be a string or an iterable object, got {type(context_uri)}"
            )

        if offset is not None:
            if can_set_offset:
                _offset: Dict[str, Union[int, str]]

                if isinstance(offset, str):
                    _offset = {"uri": offset}

                elif isinstance(offset, int):
                    _offset = {"position": offset}

                else:
                    raise TypeError(
                        f"`offset` should be either a string or an integer, got {type(offset)}"
                    )

                payload["offset"] = _offset
            else:
                raise ValueError(
                    "not able to set `offset` as either `context_uri` was not a list or it was a playlist or album uri."
                )

        if device_id is not None:
            params["device_id"] = device_id

        return self.request(route, params=params, json=payload)

    def shuffle_playback(
        self, state: bool, *, device_id: Optional[str] = None
    ) -> Awaitable:
        """Toggle shuffle on or off for user’s playback.

        Parameters
        ----------
        state : :class:`bool`
            True : Shuffle user’s playback
            False : Do not shuffle user’s playback.
        device_id : Optional[:class:`str`]
            The id of the device this command is targeting. If not supplied, the user’s currently active device is the target.
        """
        route = self.route("PUT", "/me/player/shuffle")
        payload: Dict[str, Any] = {"state": f"{bool(state)}".lower()}
        if device_id is not None:
            payload["device_id"] = device_id

        return self.request(route, params=payload)

    def transfer_player(
        self, device_id: str, *, play: Optional[bool] = False
    ) -> Awaitable:
        """Transfer playback to a new device and determine if it should start playing.

        .. note:

            Note that a value of false for the play parameter when also transferring to another device_id will not pause playback.
            To ensure that playback is paused on the new device you should send a pause command to the currently active device before transferring to the new device_id.

        Parameters
        ----------
        device_id : :class:`str`
            A Spotify Device ID
        play : Optional[:class:`bool`]
            True: ensure playback happens on new device.
            False or not provided: keep the current playback state.
        """
        route = self.route("PUT", "/me/player")
        payload: Dict[str, Any] = {"device_ids": [device_id], "play": play}

        return self.request(route, json=payload)

    def add_playlist_tracks(
        self,
        playlist_id: str,
        tracks: Sequence[Union[str]],
        position: Optional[int] = None,
    ) -> Awaitable:
        """Add one or more tracks to a user’s playlist.

        Parameters
        ----------
        playlist_id : :class:`str`
            The Spotify ID for the playlist.
        tracks : Sequence[Union[:class:`str`]]
            A sequence of track URIs.
        position : Optional[:class:`int`]
            The position to insert the tracks, a zero-based index.
        """
        route = self.route(
            "POST", "/playlists/{playlist_id}/tracks", playlist_id=playlist_id
        )

        payload: Dict[str, Any] = {"uris": [track for track in tracks]}

        if position is not None:
            payload["position"] = position

        return self.request(route, json=payload)

    def change_playlist_details(
        self,
        playlist_id: str,
        *,
        name: Optional[str] = None,
        public: Optional[bool] = None,
        collaborative: Optional[bool] = None,
        description: Optional[str] = None,
    ) -> Awaitable:
        """Change a playlist’s name and public/private state. (The user must, of course, own the playlist.)

        Parameters
        ----------
        playlist_id : :class:`str`
            The Spotify ID for the playlist.
        name : :class:`str`
            The name for the new playlist
        public : Optional[:class:`bool`]
            Defaults to true . If true the playlist will be public, if false it will be private
        collaborative : Optional[:class:`bool`]
            Defaults to false . If true the playlist will be collaborative.

            .. note::
                to create a collaborative playlist you must also set public to false
        description : Optional[:class:`str`]
            The value for playlist description as displayed in Spotify Clients and in the Web API.
        """
        route = self.route("PUT", "/playlists/{playlist_id}", playlist_id=playlist_id)

        payload: Dict[str, Any] = filter_items(
            {
                "name": name,
                "public": public,
                "collaborative": collaborative,
                "description": description,
            }
        )

        return self.request(route, json=payload)

    def create_playlist(
        self,
        user_id: str,
        *,
        name: str,
        public: Optional[bool] = True,
        collaborative: Optional[bool] = False,
        description: Optional[str] = "",
    ) -> Awaitable:
        """Create a playlist for a Spotify user. (The playlist will be empty until you add tracks.)

        Parameters
        ----------
        user_id : :class:`str`
            The user’s Spotify user ID.
        name : :class:`str`
            The name for the new playlist
        public : Optional[:class:`bool`]
            Defaults to true . If true the playlist will be public, if false it will be private
        collaborative : Optional[:class:`bool`]
            Defaults to false . If true the playlist will be collaborative.

            .. note::
                to create a collaborative playlist you must also set public to false
        description : Optional[:class:`str`]
            The value for playlist description as displayed in Spotify Clients and in the Web API.
        """
        route = self.route("POST", "/users/{user_id}/playlists", user_id=user_id)

        payload: Dict[str, Any] = {
            "name": name,
            "public": public,
            "collaborative": collaborative,
            "description": description,
        }

        return self.request(route, json=payload)

    def follow_playlist(
        self, playlist_id: str, *, public: Optional[bool] = True
    ) -> Awaitable:
        """Add the current user as a follower of a playlist.

        Parameters
        ----------
        playlist_id : :class:`str`
            The Spotify ID of the playlist. Any playlist can be followed, regardless of its public/private status, as long as you know its playlist ID.
        public : Optional[:class:`bool`]
            Defaults to true. If true the playlist will be included in user’s public playlists, if false it will remain private.
        """
        route = self.route(
            "PUT", "/playlists/{playlist_id}/followers", playlist_id=playlist_id
        )

        payload: Dict[str, Any] = {"public": public}

        return self.request(route, json=payload)

    def current_playlists(
        self, *, limit: Optional[int] = 20, offset: Optional[int] = 0
    ) -> Awaitable:
        """Get a list of the playlists owned or followed by the current Spotify user.

        Parameters
        ----------
        limit : Optional[:class:`str`]
            The maximum number of playlists to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optional[:class:`str`]
            he index of the first playlist to return. Default: 0 (the first object). Maximum offset: 100.000.
        """
        route = self.route("GET", "/me/playlists")
        return self.request(route, params={"limit": limit, "offset": offset})

    def get_playlists(
        self, user_id: str, *, limit: Optional[int] = 20, offset: Optional[int] = 0
    ) -> Awaitable:
        """Get a list of the playlists owned or followed by a Spotify user.

        Parameters
        ----------
        user_id : :class:`str`
            The user’s Spotify user ID.
        limit : Optional[:class:`str`]
            The maximum number of playlists to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optional[:class:`str`]
            he index of the first playlist to return. Default: 0 (the first object). Maximum offset: 100.000.
        """
        route = self.route("GET", "/users/{user_id}/playlists", user_id=user_id)
        return self.request(route, params={"limit": limit, "offset": offset})

    def get_playlist_cover_image(self, playlist_id: str) -> Awaitable:
        """Get the current image associated with a specific playlist.

        Parameters
        ----------
        playlist_id : :class:`str`
            The Spotify ID for the playlist.
        """
        route = self.route(
            "GET", "/playlists/{playlist_id}/images", playlist_id=playlist_id
        )
        return self.request(route)

    def get_playlist(
        self,
        playlist_id: str,
        *,
        fields: Optional[str] = None,
        market: Optional[str] = None,
    ) -> Awaitable:
        """Get a playlist owned by a Spotify user.

        Parameters
        ----------
        playlist_id : :class:`str`
            The Spotify ID for the playlist.
        fields: Optional[:class:`str`]
            Filters for the query: a comma-separated list of the fields to return.
            If omitted, all fields are returned. For example, to get just the total number of tracks and the request limit: `fields=total,limit`

            A dot separator can be used to specify non-reoccurring fields, while parentheses can be used to specify reoccurring fields within objects.
            For example, to get just the added date and user ID of the adder: `fields=items(added_at,added_by.id)`

            Use multiple parentheses to drill down into nested objects, for example: `fields=items(track(name,href,album(name,href)))`

            Fields can be excluded by prefixing them with an exclamation mark, for example: `fields=items.track.album(!external_urls,images)`
        market : Optional[:class:`str`]
            An ISO 3166-1 alpha-2 country code or the string "from_token".
            Provide this parameter if you want to apply Track Relinking.
        """
        route = self.route("GET", "/playlists/{playlist_id}", playlist_id=playlist_id)
        payload: Dict[str, Any] = {}

        if fields:
            payload["fields"] = fields

        if market:
            payload["market"] = market

        return self.request(route, params=payload)

    def get_playlist_tracks(
        self,
        playlist_id: str,
        *,
        fields: Optional[str] = None,
        market: Optional[str] = None,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0,
    ) -> Awaitable:
        """Get full details of the tracks of a playlist owned by a Spotify user.

        Parameters
        ----------
        playlist_id : :class:`str`
            The Spotify ID for the playlist.
        fields: Optional[:class:`str`]
            Filters for the query: a comma-separated list of the fields to return.
            If omitted, all fields are returned. For example, to get just the total number of tracks and the request limit: `fields=total,limit`

            A dot separator can be used to specify non-reoccurring fields, while parentheses can be used to specify reoccurring fields within objects.
            For example, to get just the added date and user ID of the adder: `fields=items(added_at,added_by.id)`

            Use multiple parentheses to drill down into nested objects, for example: `fields=items(track(name,href,album(name,href)))`

            Fields can be excluded by prefixing them with an exclamation mark, for example: `fields=items.track.album(!external_urls,images)`
        limit : Optional[:class:`str`]
            The maximum number of tracks to return. Default: 100. Minimum: 1. Maximum: 100.
        offset : Optional[:class:`str`]
            The index of the first track to return. Default: 0 (the first object).
        market : Optional[:class:`str`]
            An ISO 3166-1 alpha-2 country code or the string "from_token".
            Provide this parameter if you want to apply Track Relinking.
        """
        route = self.route(
            "GET", "/playlists/{playlist_id}/tracks", playlist_id=playlist_id
        )
        payload: Dict[str, Any] = {"limit": limit, "offset": offset}

        if fields:
            payload["fields"] = fields

        if market:
            payload["market"] = market

        return self.request(route, params=payload)

    def remove_playlist_tracks(
        self,
        playlist_id: str,
        tracks: Sequence[Union[str, Dict[str, Any]]],
        *,
        snapshot_id: str = None,
    ) -> Awaitable:
        """Remove one or more tracks from a user’s playlist.

        Parameters
        ----------
        playlist_id : str
            The id of the playlist to target
        tracks : Sequence[Union[str, Dict[str, Union[str, int]]]]
            Either a sequence of track URIs to remove a specific occurence
            of a track or for targeted removal pass in a dict that looks like
            `{'uri': URI, 'position': POSITIONS}` where `URI` is  track URI and
            `POSITIONS` is an list of integers
        snapshot_id : Optional[str]
            The snapshot to target.
        """
        route = self.route(
            "DELETE ", "/playlists/{playlist_id}/tracks", playlist_id=playlist_id
        )
        payload: Dict[str, Any] = {
            "tracks": [
                ({"uri": track} if isinstance(track, str) else track)
                for track in tracks
            ]
        }

        if snapshot_id:
            payload["snapshot_id"] = snapshot_id

        return self.request(route, json=payload)

    def reorder_playlists_tracks(
        self,
        playlist_id: str,
        range_start: int,
        range_length: int,
        insert_before: int,
        *,
        snapshot_id: Optional[str] = None,
    ) -> Awaitable:
        """Reorder a track or a group of tracks in a playlist.

        Visualization of how reordering tracks works

        .. image:: /images/visualization-reordering-tracks.png

        .. note::
            When reordering tracks, the timestamp indicating when they were added and the user who added them will be kept untouched.
            In addition, the users following the playlists won’t be notified about changes in the playlists when the tracks are reordered.

        Parameters
        ----------
        playlist_id : :class:`str`
            The Spotify ID for the playlist.
        range_start : :class:`int`
            The position of the first track to be reordered.
        range_length : :class:`int`
            The amount of tracks to be reordered. Defaults to 1 if not set.

            The range of tracks to be reordered begins from the range_start position, and includes the range_length subsequent tracks.
        insert_before : :class:`int`
            The position where the tracks should be inserted.

            To reorder the tracks to the end of the playlist, simply set insert_before to the position after the last track.
        snapshot_id : Optional[:class:`str`]
            The playlist’s snapshot ID against which you want to make the changes.
        """
        route = self.route(
            "PUT", "/playlists/{playlist_id}/tracks", playlist_id=playlist_id
        )
        payload: Dict[str, Any] = {
            "range_start": range_start,
            "range_length": range_length,
            "insert_before": insert_before,
        }

        if snapshot_id:
            payload["snapshot_id"] = snapshot_id

        return self.request(route, json=payload)

    def replace_playlist_tracks(
        self, playlist_id: str, tracks: Sequence[str]
    ) -> Awaitable:
        """Replace all the tracks in a playlist, overwriting its existing tracks.

        .. note::

            This powerful request can be useful for replacing tracks, re-ordering existing tracks, or clearing the playlist.

        Parameters
        ----------
        playlist_id : :class:`str`
            The Spotify ID for the playlist.
        tracks : Sequence[:class:`str`]
            A list of tracks to replace with.
        """
        route = self.route(
            "PUT", "/playlists/{playlist_id}/tracks", playlist_id=playlist_id
        )
        payload: Dict[str, Any] = {"uris": tuple(tracks)}

        return self.request(route, json=payload)

    def upload_playlist_cover_image(
        self, playlist_id: str, file: BinaryIO
    ) -> Awaitable:
        """Replace the image used to represent a specific playlist.

        Parameters
        ----------
        playlist_id : :class:`str`
            The Spotify ID for the playlist.
        file : File-like object
            An file-like object that supports reading
            the contents that are being read should be :class:`bytes`
        """
        route = self.route(
            "PUT", "/playlists/{playlist_id}/images", playlist_id=playlist_id
        )
        return self.request(
            route, data=b64encode(file.read()), content_type="image/jpeg"
        )

    def track_audio_analysis(self, track_id: str) -> Awaitable:
        """Get a detailed audio analysis for a single track identified by its unique Spotify ID.

        The Audio Analysis endpoint provides low-level audio analysis for all of the tracks in the Spotify catalog.
        The Audio Analysis describes the track’s structure and musical content, including rhythm, pitch, and timbre.
        All information is precise to the audio sample.

        Many elements of analysis include confidence values, a floating-point number ranging from 0.0 to 1.0.
        Confidence indicates the reliability of its corresponding attribute.
        Elements carrying a small confidence value should be considered speculative.
        There may not be sufficient data in the audio to compute the attribute with high certainty.

        Parameters
        ----------
        track_id : :class:`str`
            The Spotify ID for the track.
        """
        route = self.route("GET", "/audio-analysis/{id}", id=track_id)
        return self.request(route)

    def track_audio_features(self, track_id: str) -> Awaitable:
        """Get audio feature information for a single track identified by its unique Spotify ID.

        Parameters
        ----------
        track_id : :class:`str`
            The Spotify ID for the track.
        """
        route = self.route("GET", "/audio-features/{id}", id=track_id)
        return self.request(route)

    def audio_features(self, track_ids: List[str]) -> Awaitable:
        """Get audio features for multiple tracks based on their Spotify IDs.

        Parameters
        ----------
        track_ids : List[:class:`str`]
            A comma-separated list of the Spotify IDs for the tracks. Maximum: 100 IDs.
        """
        route = self.route("GET", "/audio-features")
        return self.request(route, params={"ids": track_ids})

    def track(self, track_id: str, market: Optional[str] = None) -> Awaitable:
        """Get Spotify catalog information for a single track identified by its unique Spotify ID.

        Parameters
        ----------
        track_id : :class:`str`
            The Spotify ID for the track.
        market : Optional[:class:`str`]
            An ISO 3166-1 alpha-2 country code or the string "from_token".
            Provide this parameter if you want to apply Track Relinking.
        """
        route = self.route("GET", "/tracks/{id}", id=track_id)
        payload: Dict[str, Any] = {}

        if market is not None:
            payload["market"] = market

        return self.request(route, params=payload)

    def tracks(self, track_ids: List[str], market: Optional[str] = None) -> Awaitable:
        """Get Spotify catalog information for multiple tracks based on their Spotify IDs.

        Parameters
        ----------
        track_ids : List[:class:`str`]
            A comma-separated list of the Spotify IDs for the tracks. Maximum: 50 IDs.
        market : Optional[:class:`str`]
            An ISO 3166-1 alpha-2 country code or the string "from_token".
            Provide this parameter if you want to apply Track Relinking.
        """
        route = self.route("GET", "/tracks")
        payload: Dict[str, Any] = {"ids": track_ids}

        if market is not None:
            payload["market"] = market

        return self.request(route, params=payload)

    def current_user(self) -> Awaitable:
        """Get detailed profile information about the current user (including the current user’s username)."""
        route = self.route("GET", "/me")
        return self.request(route)

    def user(self, user_id: str) -> Awaitable:
        """Get public profile information about a Spotify user.

        Parameters
        ---------
        user_id : class:`str`
            The user’s Spotify user ID.
        """
        route = self.route("GET", "/users/{user_id}", user_id=user_id)
        return self.request(route)

    def search(  # pylint: disable=invalid-name
        self,
        q: str,
        query_type: str = "track,playlist,artist,album",
        market: str = "US",
        limit: int = 20,
        offset: int = 0,
        include_external: Optional[str] = None,
    ) -> Awaitable:
        """Get Spotify Catalog information about artists, albums, tracks or playlists that match a keyword string.

        Parameters
        ----------
        q : :class:`str`
            Search query keywords and optional field filters and operators. e.g. `roadhouse blues.`

        query_type : Optional[:class:`str`]
            A comma-separated list of item types to search across. (default: "track,playlist,artist,album")
            Valid types are: album, artist, playlist, and track.
            Search results include hits from all the specified item types.

        market : Optional[:class:`str`]
            An ISO 3166-1 alpha-2 country code or the string "from_token". (default: "US")
            If a country code is specified, only artists, albums, and tracks with content that is playable in that market is returned.

            .. note::
                - Playlist results are not affected by the market parameter.
                - If market is set to "from_token", and a valid access token is specified in the request header, only
                    content playable in the country associated with the user account, is returned.
                - Users can view the country that is associated with their account in the account settings. A user must
                    grant access to the user-read-private scope prior to when the access token is issued.

        limit : Optional[:class:`int`]
            Maximum number of results to return. (Default: 20, Minimum: 1, Maximum: 50)

            .. note::
                The limit is applied within each type, not on the total response.
                For example, if the limit value is 3 and the type is artist,album, the response contains 3 artists and 3 albums.

        offset : Optional[:class:`int`]
            The index of the first result to return.
            Default: 0 (the first result).
            Maximum offset (including limit): 10,000.
            Use with limit to get the next page of search results.

        include_external : Optional[:class:`str`]
            Possible values: `audio`
            If `include_external=audio` is specified the response will include any relevant audio content that is hosted externally.
            By default external content is filtered out from responses.

        """
        route = self.route("GET", "/search")
        payload: Dict[str, Any] = {
            "q": quote(q),
            "type": query_type,
            "limit": limit,
            "offset": offset,
        }

        if market:
            payload["market"] = market

        if include_external is not None:
            payload["include_external"] = include_external

        return self.request(route, params=payload)


REFRESH_TOKEN_URL = "https://accounts.spotify.com/api/token?grant_type=refresh_token&refresh_token={refresh_token}"


class HTTPUserClient(HTTPClient):
    """HTTPClient for access to user endpoints."""

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        token: str = None,
        refresh_token: str = None,
        loop=None,
    ):
        assert token or refresh_token
        super().__init__(client_id, client_secret, loop=loop)
        if token:
            self.bearer_info = {"access_token": token}
        self.refresh_token = refresh_token

    async def get_bearer_info(self, *_, **__):
        if not self.refresh_token:
            # Should only happen if User.from_token didn't receive refresh_token
            raise SpotifyException(
                "Access token expired and no refresh token was provided"
            )

        headers = {
            "Authorization": f"Basic {b64encode(':'.join((self.client_id, self.client_secret)).encode()).decode()}",
            "Content-Type": "application/x-www-form-urlencoded",
        }

        route = ("POST", REFRESH_TOKEN_URL.format(refresh_token=self.refresh_token))
        return await self.request(route, headers=headers)



import discord
import typing
import random

from . import exceptions
from .node import Node

from discord.ext import commands


class NodePool:
    """The base class for the node poll. This holds all the nodes that are to be used by the bot."""
    
    _nodes: dict = {}

    def __repr__(self):
        return f"<Pomice.NodePool node_count={len(self._nodes.values())}>"

    @property
    def nodes(self):
        """Property which returns a dict with the node identifier and the Node object."""
        return self._nodes
    

    @classmethod
    def get_node(self, *, identifier: str = None) -> Node:
        """Fetches a node from the node pool using it's identifier. If no identifier is provided, it will choose a node at random."""
        available_nodes = {identifier: node for identifier, node in self._nodes.items()}
        if not available_nodes:
            raise exceptions.NoNodesAvailable('There are no nodes available.')

        if identifier is None:
            return random.choice(list(available_nodes.values()))

        return available_nodes.get(identifier, None)
     
    @classmethod
    async def create_node(self, bot: typing.Union[commands.Bot, discord.Client, commands.AutoShardedBot, discord.AutoShardedClient], host: str, port: str, password: str, identifier: str) -> Node:
        """Creates a Node object to be then added into the node pool. If you like to have Spotify searching capabilites, pass in valid Spotify API credentials."""
        if identifier in self._nodes.keys():
            raise exceptions.NodeCreationError(f"A node with identifier '{identifier}' already exists.")

        node = Node(pool=self, bot=bot, host=host, port=port, password=password, identifier=identifier)
        await node.connect()
        self._nodes[node._identifier] = node
        return node



    


from os import strerror
import aiohttp
import discord
import asyncio
import typing
import json
import socket
import time
import re

from discord.ext import commands
from typing import Optional, Union
from urllib.parse import quote
from . import spotify
from . import events
from . import exceptions
from . import objects
from . import __version__
from .utils import ExponentialBackoff, NodeStats


SPOTIFY_URL_REGEX = re.compile(r'https?://open.spotify.com/(?P<type>album|playlist|track)/(?P<id>[a-zA-Z0-9]+)')

class Node:
    """The base class for a node. 
    This node object represents a Lavalink node. 
    If you want to enable Spotify searching, pass in a proper Spotify Client ID and Spotify Client Secret"""
    def __init__(self, pool, bot: Union[commands.Bot, discord.Client, commands.AutoShardedBot, discord.AutoShardedClient], host: str, port: int, password: str, identifier: str, spotify_client_id: Optional[str], spotify_client_secret: Optional[str]):
        self._bot = bot
        self._host = host
        self._port = port
        self._password = password
        self._identifier = identifier
        self._pool = pool

        self._websocket_uri = f"ws://{self._host}:{self._port}"
        self._rest_uri = f"http://{self._host}:{self._port}"

        self._session = aiohttp.ClientSession()
        self._websocket: aiohttp.ClientWebSocketResponse = None
        self._task: asyncio.Task = None

        self._connection_id = None
        self._metadata = None
        self._available = None

        self._headers = {
            "Authorization": self._password,
            "User-Id": str(self._bot.user.id),
            "Client-Name": f"Pomice/{__version__}"
        }

        self._players = {}

        self._spotify_client_id: str = spotify_client_id
        self._spotify_client_secret: str = spotify_client_secret

        if self._spotify_client_id and self._spotify_client_secret:
            self._spotify_client: spotify.Client = spotify.Client(self._spotify_client_id, self._spotify_client_secret)
            self._spotify_http_client: spotify.HTTPClient = spotify.HTTPClient(self._spotify_client_id, self._spotify_client_secret)

        self._bot.add_listener(self._update_handler, "on_socket_response")

    def __repr__(self):
        return f"<Pomice.node ws_uri={self._websocket_uri} rest_uri={self._rest_uri} player_count={len(self._players)}>"

    @property
    def is_connected(self) -> bool:
        """"Property which returns whether this node is connected or not"""
        return self._websocket is not None and not self._websocket.closed

    @property
    async def latency(self):
        """Property which returns the latency of the node in milliseconds"""
        start_time = time.time()
        await self.send(op="ping")
        end_time = await self._bot.wait_for(f"node_ping")
        return (end_time - start_time) * 1000

    @property
    async def stats(self):
        """Property which returns the node stats at any given time. 
        Typically, accessing this property is rare due to the fact that Lavalink automatically sends updated node stats every minutes."""
        await self.send(op="get-stats")
        node_stats = await self._bot.wait_for(f"node_stats")
        return node_stats

    @property
    def players(self):
        """Property which returns a dict containing the guild ID and the player object."""
        return self._players

    @property
    def bot(self):
        """Property which returns the discord.py Bot object linked to this node"""
        return self._bot

    @property
    def pool(self):
        """Property which returns the node pool this node is apart of."""
        return self._pool

    async def _update_handler(self, data: dict):
        await self._bot.wait_until_ready()

        if not data:
            return


        if data["t"] == "VOICE_SERVER_UPDATE":

            guild_id = int(data["d"]["guild_id"])
            try:
                player = self._players[guild_id]
                await player._voice_server_update(data["d"])
            except KeyError:
                return

        elif data["t"] == "VOICE_STATE_UPDATE":

            if int(data["d"]["user_id"]) != self._bot.user.id:
                return

            guild_id = int(data["d"]["guild_id"])
            try:
                player = self._players[guild_id]
                await player._voice_state_update(data["d"])
            except KeyError:
                return

        else:
            return

    async def _listen(self):
        backoff = ExponentialBackoff(base=7)

        while True:
            msg = await self._websocket.receive()
            if msg.type == aiohttp.WSMsgType.CLOSED:
                retry = backoff.delay()
                await asyncio.sleep(retry)

                if not self.is_connected:
                    self._bot.loop.create_task(self.connect())
            else:
                self._bot.loop.create_task(self._handle_payload(msg.json()))

    async def _handle_payload(self, data: dict) -> None:
        op = data.get('op', None)
        if not op:
            return

        if op == 'stats':
            self._stats = NodeStats(data)
            return

        if not (player := self._players.get(int(data['guildId']))):
            return

        if op == 'event':
            await player._dispatch_event(data)
        elif op == 'playerUpdate':
            await player._update_state(data)


    async def send(self, **data):
        if not self.available:
            raise exceptions.NodeNotAvailable(f"The node '{self.identifier}' is not currently available.")

        await self._websocket.send_str(json.dumps(data))

    def get_player(self, guild_id: int):
        """Takes a guild ID as a parameter. Returns a pomice Player object."""
        return self._players.get(guild_id, None)

    async def connect(self): 
        """Initiates a connection with a Lavalink node and adds it to the node pool."""
        await self._bot.wait_until_ready()

        try:
            self._websocket = await self._session.ws_connect(self._websocket_uri, headers=self._headers, heartbeat=60)
            self._task = self._bot.loop.create_task(self._listen())
            self._pool._nodes[self._identifier] = self
            self.available = True
            return self

        except aiohttp.WSServerHandshakeError:
            raise exceptions.NodeConnectionFailure(f"The password for node '{self.identifier}' is invalid.")
        except aiohttp.InvalidURL:
            raise exceptions.NodeConnectionFailure(f"The URI for node '{self.identifier}' is invalid.")
        except socket.gaierror:
            raise exceptions.NodeConnectionFailure(f"The node '{self.identifier}' failed to connect.")

    async def disconnect(self):
        """Disconnects a connected Lavalink node and removes it from the node pool. This also destroys any players connected to the node."""
        for player in self.players.copy().values():
            await player.destroy()

        await self._websocket.close()
        del self._pool.nodes[self._identifier]
        self.available = False
        self._task.cancel()
        
    async def get_tracks(self, query: str, ctx: commands.Context = None):
        """Fetches tracks from the node's REST api to parse into Lavalink.
        If you passed in Spotify API credentials, you can also pass in a Spotify URL of a playlist, album or track
        and it will be parsed accordingly.
        You can also pass in a discord.py Context object to get a Context object on any track you search.
        """

        if spotify_url_check := SPOTIFY_URL_REGEX.match(query):
            
            if not self._spotify_client_id and not self._spotify_client_secret:
                raise exceptions.InvalidSpotifyClientAuthorization("You did not provide proper Spotify client authorization credentials. If you would like to use the Spotify searching feature, please obtain Spotify API credentials here: https://developer.spotify.com/")

            search_type = spotify_url_check.group('type')
            spotify_id = spotify_url_check.group('id')
            if search_type == "playlist":
                results: spotify.Playlist = spotify.Playlist(client=self._spotify_client, data=await self._spotify_http_client.get_playlist(spotify_id))
                try:
                    search_tracks = await results.get_all_tracks()
                    tracks = [
                        objects.Track(
                                track_id='spotify',
                                ctx=ctx,
                                info={'title': track.name or 'Unknown', 'author': ', '.join(artist.name for artist in track.artists) or 'Unknown',
                                            'length': track.duration or 0, 'identifier': track.id or 'Unknown', 'uri': track.url or 'spotify',
                                            'isStream': False, 'isSeekable': False, 'position': 0, 'thumbnail': track.images[0].url if track.images else None},
                                
                        ) for track in search_tracks
                    ]
                    return objects.Playlist(playlist_info={"name": results.name, "selectedTrack": search_tracks[0]}, tracks=tracks, ctx=ctx)
                except:
                    raise exceptions.SpotifyPlaylistLoadFailed(f"Unable to find results for {query}")
            elif search_type == "album":
                results: spotify.Album = await self._spotify_client.get_album(spotify_id=spotify_id)
                try:
                    search_tracks = await results.get_all_tracks()
                    tracks = [
                        objects.Track(
                                track_id='spotify',
                                ctx=ctx,
                                info={'title': track.name or 'Unknown', 'author': ', '.join(artist.name for artist in track.artists) or 'Unknown',
                                            'length': track.duration or 0, 'identifier': track.id or 'Unknown', 'uri': track.url or 'spotify',
                                            'isStream': False, 'isSeekable': False, 'position': 0, 'thumbnail': track.images[0].url if track.images else None},
                                
                        ) for track in search_tracks
                    ]
                    
                    return objects.Playlist(playlist_info={"name": results.name, "selectedTrack": search_tracks[0]}, tracks=tracks, ctx=ctx)
                except:
                    raise exceptions.SpotifyAlbumLoadFailed(f"Unable to find results for {query}")
            elif search_type == 'track':
                try:
                    results: spotify.Track = await self._spotify_client.get_track(spotify_id=spotify_id)
                    return objects.Track(
                            track_id='spotify',
                            ctx=ctx,
                            info={'title': results.name or 'Unknown', 'author': ', '.join(artist.name for artist in results.artists) or 'Unknown',
                                        'length': results.duration or 0, 'identifier': results.id or 'Unknown', 'uri': results.url or 'spotify',
                                        'isStream': False, 'isSeekable': False, 'position': 0, 'thumbnail': results.images[0].url if results.images else None},)
                except:
                    raise exceptions.SpotifyTrackLoadFailed(f"Unable to find results for {query}")
 
        
        else:
            async with self._session.get(url=f"{self._rest_uri}/loadtracks?identifier={quote(query)}", headers={"Authorization": self._password}) as response:
                data = await response.json()

        load_type = data.get("loadType")

        if not load_type:
            raise exceptions.TrackLoadError("There was an error while trying to load this track.")

        elif load_type == "LOAD_FAILED":
            raise exceptions.TrackLoadError(f"There was an error of severity '{data['severity']}' while loading tracks.\n\n{data['cause']}")

        elif load_type == "NO_MATCHES":
            return None

        elif load_type == "PLAYLIST_LOADED":
            return objects.Playlist(playlist_info=data["playlistInfo"], tracks=data["tracks"], ctx=ctx)
            
        elif load_type == "SEARCH_RESULT" or load_type == "TRACK_LOADED":
            return [objects.Track(track_id=track["track"], info=track["info"], ctx=ctx) for track in data["tracks"]]
          
                



from typing import Optional, Callable, Type

import spotify


class SpotifyBase:
    """The base class all Spotify models **must** derive from.

    This base class is used to transparently construct spotify
    models based on the :class:`spotify,Client` type.

    Currently it is used to detect whether a Client is a synchronous
    client and, if as such, construct and return the appropriate
    synchronous model.
    """

    __slots__ = ()

    def __new__(cls, client, *_, **__):
        if not isinstance(client, spotify.Client):
            raise TypeError(
                f"{cls!r}: expected client argument to be an instance of `spotify.Client`. Instead got {type(client)}"
            )

        if hasattr(client, "__client_thread__"):
            cls = getattr(  # pylint: disable=self-cls-assignment
                spotify.sync.models, cls.__name__
            )

        return object.__new__(cls)

    async def from_href(self):
        """Get the full object from spotify with a `href` attribute.

        .. note ::

            This can be used to get an updated model of the object.

        Returns
        -------
        model : SpotifyBase
            An instance of whatever the class was before.

        Raises
        ------
        TypeError
            This is raised if the model has no `href` attribute.

            Additionally if the model has no `http` attribute and
            the model has no way to access its client, while theoretically
            impossible its a failsafe, this will be raised.
        """
        if not hasattr(self, "href"):
            raise TypeError(
                "Spotify object has no `href` attribute, therefore cannot be retrived"
            )

        if hasattr(self, "http"):
            return await self.http.request(  # pylint: disable=no-member
                ("GET", self.href)  # pylint: disable=no-member
            )

        klass = type(self)

        try:
            client = getattr(self, f"_{klass.__name__}__client")
        except AttributeError:
            raise TypeError("Spotify object has no way to access a HTTPClient.")
        else:
            http = client.http  # pylint: disable=no-member

        data = await http.request(("GET", self.href))  # pylint: disable=no-member

        return klass(client, data)


class URIBase(SpotifyBase):
    """Base class used for inheriting magic methods for models who have URIs.

    This class inherits from :class:`SpotifyBase` and is used to reduce boilerplate
    in spotify models by supplying a `__eq__`, `__ne__`, and `__str__` double underscore
    methods.

    The objects that inherit from :class:`URIBase` support equality and string casting.

     - Two objects are equal if **They are strictly the same type and have the same uri**
     - Casting to a string will return the uri of the object.
    """

    uri = repr(None)

    def __hash__(self):
        return hash(self.uri)  # pylint: disable=no-member

    def __eq__(self, other):
        return (
            type(self) is type(other) and self.uri == other.uri
        )  # pylint: disable=no-member

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.uri  # pylint: disable=no-member


class AsyncIterable(SpotifyBase):
    """Base class intended for all models that can be asynchronously iterated over.

    This class implements two magic class vars:

     * `__aiter_fetch__` ~ A coroutine function that accepts a keyword argument named `option`
     * `__aiter_klass__` ~ A spotify model class, essentially a type that subclasses `SpotifyBase`

    Additionally the class implements `__aiter__` that will exhaust the paging
    objects returned by the `__aiter_fetch__` calls and yield each data item in
    said paging objects as an instance of `__aiter_klass__`.
    """

    __aiter_fetch__: Optional[Callable] = None
    __aiter_klass__: Optional[Type[SpotifyBase]] = None

    async def __aiter__(self):
        client = getattr(self, f"_{type(self).__name__}__client")

        assert self.__aiter_fetch__ is not None
        fetch = self.__aiter_fetch__

        assert self.__aiter_klass__ is not None
        klass = self.__aiter_klass__

        total = None
        processed = offset = 0

        while total is None or processed < total:
            data = await fetch(offset=offset)  # pylint: disable=not-callable

            if total is None:
                assert "total" in data
                total = data["total"]

            assert "items" in data
            for item in data["items"]:
                processed += 1
                yield klass(client, item)  # pylint: disable=not-callable

            offset += 50



import setuptools

with open("README.md") as f:
    readme = f.read()


setuptools.setup(
    name="pomice",
    author="cloudwithax",
    version="1.0.3",
    url="https://github.com/cloudwithax/pomice",
    packages=setuptools.find_packages(),
    license="GPL",
    description="The modern Lavalink wrapper designed for Discord.py",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=['discord.py>=1.7.1'],
    extra_require=None,
    classifiers=[
        "Framework :: AsyncIO",
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'Intended Audience :: Developers',
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries',
        "Topic :: Internet"
    ],
    python_requires='>=3.8',
    keywords=['pomice', 'lavalink', "discord.py"],
)


from asyncio import new_event_loop, run_coroutine_threadsafe, set_event_loop
from threading import Thread, RLock, get_ident
from typing import Any, Coroutine


class EventLoopThread(Thread):
    """A surrogate thread that spins an asyncio event loop."""

    def __init__(self):
        super().__init__(daemon=True)

        self.__lock = RLock()
        self.__loop = loop = new_event_loop()
        loop.__spotify_thread__ = self

    # Properties

    @property
    def loop(self):
        return self.__loop

    # Overloads

    def run(self):
        set_event_loop(self.__loop)
        self.__loop.run_forever()

    # Public API

    def run_coroutine_threadsafe(self, coro: Coroutine) -> Any:
        """Like :func:`asyncio.run_coroutine_threadsafe` but for this specific thread."""

        # If the current thread is the same
        # as the event loop Thread.
        #
        # then we're in the process of making
        # nested calls to await other coroutines
        # and should pass back the coroutine as it should be.
        if get_ident() == self.ident:
            return coro

        # Double lock because I haven't looked
        # into whether this deadlocks under whatever
        # conditions, Best to play it safe.
        with self.__lock:
            future = run_coroutine_threadsafe(coro, self.__loop)

        return future.result()
